#
#-------------------------------------------------------------
#   Copyright 2004-2009 Synopsys, Inc.
#   Copyright 2010-2011 Mentor Graphics Corporation
#   Copyright 2010-2011 Cadence Design Systems, Inc.
#   Copyright 2013 Semifore, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#    TODO add modifications
#-------------------------------------------------------------

from typing import Dict

from ..base.uvm_object import UVMObject
from ..base.uvm_globals import uvm_report_error, uvm_report_warning
from ..base.uvm_resource_db import UVMResourceDb
from ..macros.uvm_object_defines import uvm_object_utils
from ..macros.uvm_message_defines import uvm_error
from .uvm_reg_model import (UVM_CHECK, UVM_FRONTDOOR, UVM_IS_OK, UVM_NOT_OK, UVM_NO_CHECK,
                            UVM_NO_COVERAGE, UVM_PREDICT, UVM_PREDICT_DIRECT, UVM_PREDICT_READ,
                            UVM_PREDICT_WRITE, uvm_fatal)
from .uvm_reg_item import UVMRegItem
from .uvm_reg_cbs import UVMRegFieldCbIter

NO_RAND_SET = {"RO", "RC", "RS", "WC", "WS",
        "W1C", "W1S", "W1T", "W0C", "W0S", "W0T",
        "W1SRC", "W1CRS", "W0SRC", "W0CRS", "WSRC", "WCRS",
        "WOC", "WOS"}

KNOWN_ACCESSES = ["RO", "RW", "RC", "RS", "WC", "WS", "W1C", "W1S", "W1T",
        "W0C", "W0S", "W0T", "WRC", "WRS", "W1SRC", "W1CRS", "W0SRC", "W0CRS",
        "WSRC", "WCRS", "WO", "WOC", "WOS", "W1", "WO1"]

#constraint uvm_reg_field_valid {
#   if (`UVM_REG_DATA_WIDTH > self.m_size) {
#      value < (`UVM_REG_DATA_WIDTH'h1 << self.m_size)
#   }
#}


class UVMRegField(UVMObject):
    """
    Field abstraction class

    A field represents a set of bits that behave consistently
    as a single entity.

    A field is contained within a single register, but may
    have different access policies depending on the address map
    use the access the register (thus the field).
    """

    m_max_size = 0
    m_policy_names = {}  # type: Dict[str, int]
    m_predefined = False



    def __init__(self, name='uvm_reg_field'):
        """
        This method should not be used directly.
        The `UVMRegField.type_id.create()` factory method
        should be used instead.

        Args:
            name: (str): Name of the register field
        """
        UVMObject.__init__(self, name)
        self.value = 0  # Mirrored after randomize()
        self.m_mirrored = 0  # What we think is in the HW
        self.m_desired = 0  # Mirrored after set()
        self.m_access = ""
        self.m_parent = None  # uvm_reg
        self.m_lsb = 0
        self.m_size = 0
        self.m_volatile = False
        self.m_reset = {}  # string -> uvm_reg_data_t
        self.m_written = False
        self.m_read_in_progress = False
        self.m_write_in_progress = False
        self.m_fname = ""
        self.m_lineno = 0
        self.m_cover_on = 0
        self.m_individually_accessible = False
        self.m_check = UVM_CHECK  # uvm_check_e
        if UVMRegField.m_predefined is False:
            UVMRegField.m_predefined = UVMRegField.m_predefine_policies()


    def configure(self, parent, size, lsb_pos, access, volatile, reset, has_reset, is_rand,
            individually_accessible):
        """
        Instance-specific configuration

        Specify the `parent` register of this field, its
        `size` in bits, the position of its least-significant bit
        within the register relative to the least-significant bit
        of the register, its `access` policy, volatility,
        "HARD" `reset` value,
        whether the field value is actually reset
        (the `reset` value is ignored if `FALSE`),
        whether the field value may be randomized and
        whether the field is the only one to occupy a byte lane in the register.

        See `set_access` for a specification of the pre-defined
        field access policies.

        If the field access policy is a pre-defined policy and NOT one of
        "RW", "WRC", "WRS", "WO", "W1", or "WO1",
        the value of `is_rand` is ignored and the rand_mode() for the
        field instance is turned off since it cannot be written.

        Args:
            parent (UVMReg): Parent register.
            size (int): Size of the field in bits
            lsb_pos (int): LSB position of the field.
            access (str): Access of the fields (ie "RW" or "RO")
            volatile (bool):
            reset:
            has_reset (bool):
            is_rand (bool):
            individually_accessible (bool)
        """
        self.m_parent = parent
        if size == 0:
            uvm_report_error("RegModel",
               "Field \"{}\" cannot have 0 bits".format(self.get_full_name()))
            size = 1

        self.m_size      = size
        self.rand('value', range(1 << size))
        self.m_volatile  = volatile
        self.m_access    = access.upper()
        self.m_lsb       = lsb_pos
        self.m_cover_on  = UVM_NO_COVERAGE
        self.m_written   = False
        if volatile:
            self.m_check = UVM_NO_CHECK
        else:
            self.m_check = self.m_check
        self.m_individually_accessible = individually_accessible

        if has_reset:
            self.set_reset(reset)
        else:
            UVMResourceDb.set("REG::" + self.get_full_name(),
                   "NO_REG_HW_RESET_TEST", 1)

        self.m_parent.add_field(self)

        if not (self.m_access in UVMRegField.m_policy_names):
            uvm_report_error("RegModel", ("Access policy '" + access
             + "' for field '" + self.get_full_name() + "' is not defined.  Setting to RW"))
            self.m_access = "RW"

        if size > UVMRegField.m_max_size:
            UVMRegField.m_max_size = size

        # Ignore is_rand if the field is known not to be writeable
        # i.e. not "RW", "WRC", "WRS", "WO", "W1", "WO1"
        if access in NO_RAND_SET:
            self.is_rand = False
        #if not is_rand:
        #    self.value.rand_mode(0)

    #---------------------
    # Group: Introspection
    #---------------------

    def get_full_name(self):
        """
        Get the hierarchical name

        Return the hierarchal name of this field
        The base of the hierarchical name is the root block.

        Returns:
            str: Full hier name (including parent's name)
        """
        return self.m_parent.get_full_name() + "." + self.get_name()

    def get_parent(self):
        """
        Get the parent register

        Returns:
            UVMReg: Parent register of this field
        """
        return self.m_parent

    def get_register(self):
        """
        Get the parent register

        Returns:
            UVMReg: Parent register of this field
        """
        return self.m_parent

    def get_lsb_pos(self):
        """
        Return the position of the field

        Returns the index of the least significant bit of the field
        in the register that instantiates it.
        An offset of 0 indicates a field that is aligned with the
        least-significant bit of the register.

        Returns:
            int: LSB position of this field
        """
        return self.m_lsb

    def get_n_bits(self):
        """
           Returns the width, in number of bits, of the field.

        Returns:
            int: Width of the field in bits
        """
        return self.m_size

    @classmethod
    def get_max_size(cls):
        """
        Function: get_max_size
        Returns the width, in number of bits, of the largest field.

        Returns:
            int: Width of largest field
        """
        return UVMRegField.m_max_size


    def set_access(self, mode):
        """
        Modify the access policy of the field

        Modify the access policy of the field to the specified one and
        return the previous access policy.

        The pre-defined access policies are as follows.
        The effect of a read operation are applied after the current
        value of the field is sampled.
        The read operation will return the current value,
        not the value affected by the read operation (if any)::

            "RO"       - W: no effect, R: no effect
            "RW"       - W: as-is, R: no effect
            "RC"       - W: no effect, R: clears all bits
            "RS"       - W: no effect, R: sets all bits
            "WRC"      - W: as-is, R: clears all bits
            "WRS"      - W: as-is, R: sets all bits
            "WC"       - W: clears all bits, R: no effect
            "WS"       - W: sets all bits, R: no effect
            "WSRC"     - W: sets all bits, R: clears all bits
            "WCRS"     - W: clears all bits, R: sets all bits
            "W1C"      - W: 1/0 clears/no effect on matching bit, R: no effect
            "W1S"      - W: 1/0 sets/no effect on matching bit, R: no effect
            "W1T"      - W: 1/0 toggles/no effect on matching bit, R: no effect
            "W0C"      - W: 1/0 no effect on/clears matching bit, R: no effect
            "W0S"      - W: 1/0 no effect on/sets matching bit, R: no effect
            "W0T"      - W: 1/0 no effect on/toggles matching bit, R: no effect
            "W1SRC"    - W: 1/0 sets/no effect on matching bit, R: clears all bits
            "W1CRS"    - W: 1/0 clears/no effect on matching bit, R: sets all bits
            "W0SRC"    - W: 1/0 no effect on/sets matching bit, R: clears all bits
            "W0CRS"    - W: 1/0 no effect on/clears matching bit, R: sets all bits
            "WO"       - W: as-is, R: error
            "WOC"      - W: clears all bits, R: error
            "WOS"      - W: sets all bits, R: error
            "W1"       - W: 1st W after ``HARD`` reset is as-is, other W have no effects, R: no effect
            "WO1"      - W: 1st W after ``HARD`` reset is as-is, other W have no effects, R: error
            "NOACCESS" - W: no effect, R: no effect

        It is important to remember that modifying the access of a field
        will make the register model diverge from the specification
        that was used to create it.

        Args:
            mode (str): Mode from the list above.
        Returns:
        """
        set_access = self.m_access
        self.m_access = mode.upper()
        if self.m_access not in UVMRegField.m_policy_names:
            uvm_error("RegModel", "Access policy '" + self.m_access
                + "' is not a defined field access policy")
            self.m_access = set_access
        return set_access


    @classmethod
    def define_access(cls, name):
        """
        Define a new access policy value

        Because field access policies are specified using string values,
        there is no way for SystemVerilog to verify if a specific access
        value is valid or not.
        To help catch typing errors, user-defined access values
        must be defined using this method to avoid begin reported as an
        invalid access policy.

        The name of field access policies are always converted to all uppercase.

        Returns TRUE if the new access policy was not previously
        defined.
        Returns FALSE otherwise but does not issue an error message.

        Args:
            cls:
            name:
        Returns:
            bool: False if name already exists, True on success.
        """
        if not UVMRegField.m_predefined:
            UVMRegField.m_predefined = UVMRegField.m_predefine_policies()

        name = name.upper()
        if name in UVMRegField.m_policy_names:
            return False

        UVMRegField.m_policy_names[name] = 1
        return True

    @classmethod
    def m_predefine_policies(cls):
        """
        Internal function which creates the predefines policies.
        """
        if UVMRegField.m_predefined is True:
            return True
        UVMRegField.m_predefined = True
        UVMRegField.define_access("RO")
        UVMRegField.define_access("RW")
        UVMRegField.define_access("RC")
        UVMRegField.define_access("RS")
        UVMRegField.define_access("WRC")
        UVMRegField.define_access("WRS")
        UVMRegField.define_access("WC")
        UVMRegField.define_access("WS")
        UVMRegField.define_access("WSRC")
        UVMRegField.define_access("WCRS")
        UVMRegField.define_access("W1C")
        UVMRegField.define_access("W1S")
        UVMRegField.define_access("W1T")
        UVMRegField.define_access("W0C")
        UVMRegField.define_access("W0S")
        UVMRegField.define_access("W0T")
        UVMRegField.define_access("W1SRC")
        UVMRegField.define_access("W1CRS")
        UVMRegField.define_access("W0SRC")
        UVMRegField.define_access("W0CRS")
        UVMRegField.define_access("WO")
        UVMRegField.define_access("WOC")
        UVMRegField.define_access("WOS")
        UVMRegField.define_access("W1")
        UVMRegField.define_access("WO1")
        return True

    def get_access(self, reg_map=None):
        """
           Get the access policy of the field

           Returns the current access policy of the field
           when written and read through the specified address `map`.
           If the register containing the field is mapped in multiple
           address map, an address map must be specified.
           The access policy of a field from a specific
           address map may be restricted by the register's access policy in that
           address map.
           For example, a RW field may only be writable through one of
           the address maps and read-only through all of the other maps.
           If the field access contradicts the map's access value
           (field access of WO, and map access value of RO, etc), the
           method's return value is NOACCESS.
        Args:
            reg_map:
        Returns:
        """
        field_access = self.m_access
        from .uvm_reg_map import UVMRegMap
        if reg_map == UVMRegMap.backdoor():
            return field_access

        # Is the register restricted in this reg_map?
        rights = self.m_parent.get_rights(reg_map)
        if rights == "RW":
            # No restrictions
            return field_access
        elif rights == "RO":
            if field_access in ["RW", "RO", "WC", "WS",
                    "W1C", "W1S", "W1T", "W0C", "W0S", "W0T", "W1"]:
                field_access = "RO"
            elif field_access in ["RC", "WRC", "W1SRC", "W0SRC",
                    "WSRC"]:
                field_access = "RC"
            elif field_access in ["RS", "WRS", "W1CRS", "W0CRS", "WCRS"]:
                field_access = "RS"
            elif field_access in ["WO", "WOC", "WOS", "WO1"]:
                field_access = "NOACCESS"

        elif rights == "WO":
            field_access = "NOACCESS"
            if field_access == 'RW' or field_access == 'WO':
                field_access = "WO"
        else:
            field_access = "NOACCESS"
            fname = ""
            if reg_map is not None:
                fname = reg_map.get_full_name()
            uvm_report_warning("RegModel", ("Register '" + self.m_parent.get_full_name()
                + "' containing field '" + self.get_name() + "' is mapped in reg_map '"
                + fname + "' with unknown access right '"
                + self.m_parent.get_rights(reg_map) + "'"))
        return field_access

    def is_known_access(self, _map=None):
        """
           Check if access policy is a built-in one.

           Returns TRUE if the current access policy of the field,
           when written and read through the specified address `map`,
           is a built-in access policy.

        Args:
            _map:
        Returns:
        """
        acc = self.get_access(_map)
        if acc in KNOWN_ACCESSES:
            return 1
        return 0


    def set_volatility(self, volatile):
        """
           Function: set_volatility
           Modify the volatility of the field to the specified one.

           It is important to remember that modifying the volatility of a field
           will make the register model diverge from the specification
           that was used to create it.

        Args:
            volatile:
        """
        self.m_volatile = volatile


    def is_volatile(self):
        """
           Function: is_volatile
           Indicates if the field value is volatile

           UVM uses the IEEE 1685-2009 IP-XACT definition of "volatility".
           If TRUE, the value of the register is not predictable because it
           may change between consecutive accesses.
           This typically indicates a field whose value is updated by the DUT.
           The nature or cause of the change is not specified.
           If FALSE, the value of the register is not modified between
           consecutive accesses.

        Returns:
            bool: True if field volatile (see the definition above)
        """
        return self.m_volatile


    def set(self, value, fname="", lineno=0):
        """
          --------------
           Group: Access
          --------------


           Set the desired value for this field

           It sets the desired value of the field to the specified `value`
           modified by the field access policy.
           It does not actually set the value of the field in the design,
           only the desired value in the abstraction class.
           Use the `UVMReg.update` method to update the actual register
           with the desired value or the `UVMRegField.write` method
           to actually write the field and update its mirrored value.

           The final desired value in the mirror is a function of the field access
           policy and the set value, just like a normal physical write operation
           to the corresponding bits in the hardware.
           As such, this method (when eventually followed by a call to
           `UVMReg.update`)
           is a zero-time functional replacement for the `UVMRegField.write`
           method.
           For example, the desired value of a read-only field is not modified
           by this method and the desired value of a write-once field can only
           be set if the field has not yet been
           written to using a physical (for example, front-door) write operation.

           Use the `UVMRegField.predict` to modify the mirrored value of
           the field.
           Function: set

        Args:
            value:
            fname:
            lineno:
        """
        mask = (1 << self.m_size)-1
        self.m_fname = fname
        self.m_lineno = lineno
        if value >> self.m_size:
            uvm_report_warning("RegModel",
                "Specified value (0x{}) greater than field \"{}\" size ({} bits)".format(
                     value, self.get_name(), self.m_size))
            value = value & mask

        if self.m_parent.is_busy():
            uvm_report_warning("UVM/FLD/SET/BSY", ("Setting the value of field \"{}\" while containing"
                + "register \"{}\" is being accessed may result in loss of desired field value. "
                + "A race condition between threads concurrently accessing the register model is "
                + "the likely cause of the problem.").format(
                    self.get_name(), self.m_parent.get_full_name()))

        if self.m_access == "RO":
            self.m_desired = self.m_desired
        elif self.m_access == "RW":
            self.m_desired = value
        elif self.m_access == "RC":
            self.m_desired = self.m_desired
        elif self.m_access == "RS":
            self.m_desired = self.m_desired
        elif self.m_access == "WC":
            self.m_desired = 0
        elif self.m_access == "WS":
            self.m_desired = mask
        elif self.m_access == "WRC":
            self.m_desired = value
        elif self.m_access == "WRS":
            self.m_desired = value
        elif self.m_access == "WSRC":
            self.m_desired = mask
        elif self.m_access == "WCRS":
            self.m_desired = 0
        elif self.m_access == "W1C":
            self.m_desired = self.m_desired & (~value)
        elif self.m_access == "W1S":
            self.m_desired = self.m_desired | value
        elif self.m_access == "W1T":
            self.m_desired = self.m_desired ^ value
        elif self.m_access == "W0C":
            self.m_desired = self.m_desired & value
        elif self.m_access == "W0S":
            self.m_desired = self.m_desired | (~value & mask)
        elif self.m_access == "W0T":
            self.m_desired = self.m_desired ^ (~value & mask)
        elif self.m_access == "W1SRC":
            self.m_desired = self.m_desired | value
        elif self.m_access == "W1CRS":
            self.m_desired = self.m_desired & (~value)
        elif self.m_access == "W0SRC":
            self.m_desired = self.m_desired | (~value & mask)
        elif self.m_access == "W0CRS":
            self.m_desired = self.m_desired & value
        elif self.m_access == "WO":
            self.m_desired = value
        elif self.m_access == "WOC":
            self.m_desired = 0
        elif self.m_access == "WOS":
            self.m_desired = mask
        elif self.m_access == "W1":
            if self.m_written:
                self.m_desired = self.m_desired
            else:
                self.m_desired = value
        elif self.m_access == "WO1":
            if self.m_written:
                self.m_desired = self.m_desired
            else:
                self.m_desired = value
        else:
            self.m_desired = value
        self.value = self.m_desired


    def get(self, fname="", lineno=0):
        """
           Function: get

           Return the desired value of the field

           It does not actually read the value
           of the field in the design, only the desired value
           in the abstraction class. Unless set to a different value
           using the `UVMRegField.set`, the desired value
           and the mirrored value are identical.

           Use the `UVMRegField.read` or `UVMRegField.peek`
           method to get the actual field value.

           If the field is write-only, the desired/mirrored
           value is the value last written and assumed
           to reside in the bits implementing it.
           Although a physical read operation would something different,
           the returned value is the actual content.
        Args:
            fname:
            lineno:
        Returns:
        """
        self.m_fname = fname
        self.m_lineno = lineno
        return self.m_desired

    def get_mirrored_value(self, fname="", lineno=0):
        """

           Function: get_mirrored_value

           Return the mirrored value of the field

           It does not actually read the value of the field in the design, only the mirrored value
           in the abstraction class.

           If the field is write-only, the desired/mirrored
           value is the value last written and assumed
           to reside in the bits implementing it.
           Although a physical read operation would something different,
           the returned value is the actual content.

        Args:
            fname:
            lineno:
        Returns:
        """
        self.m_fname = fname
        self.m_lineno = lineno
        return self.m_mirrored

    def reset(self, kind = "HARD"):
        """
           Function: reset

           Reset the desired/mirrored value for this field.

           It sets the desired and mirror value of the field
           to the reset event specified by `kind`.
           If the field does not have a reset value specified for the
           specified reset `kind` the field is unchanged.

           It does not actually reset the value of the field in the design,
           only the value mirrored in the field abstraction class.

           Write-once fields can be modified after
           a "HARD" reset operation.

        Args:
            kind:
        """
        if not kind in self.m_reset:
            return
        self.m_mirrored = self.m_reset[kind]
        self.m_desired  = self.m_mirrored
        self.value      = self.m_mirrored
        if kind == "HARD":
            self.m_written  = False

    def get_reset(self, kind="HARD"):
        """

           Function: get_reset

           Get the specified reset value for this field

           Return the reset value for this field
           for the specified reset `kind`.
           Returns the current field value is no reset value has been
           specified for the specified reset event.

        Args:
            kind:
        Returns:
        """
        if not kind in self.m_reset:
            return self.m_desired
        return self.m_reset[kind]


    def has_reset(self, kind="HARD", delete=False):
        """
           Function: has_reset

           Check if the field has a reset value specified

           Return TRUE if this field has a reset value specified
           for the specified reset `kind`.
           If `delete` is TRUE, removes the reset value, if any.

        Args:
            kind:
            delete:
        Returns:
        """
        if kind not in self.m_reset:
            return False
        if delete:
            del self.m_reset[kind]
        return True

    def set_reset(self, value,kind="HARD"):
        """
           Function: set_reset

           Specify or modify the reset value for this field

           Specify or modify the reset value for this field corresponding
           to the cause specified by `kind`.

        Args:
            value:
            kind:
        """
        self.m_reset[kind] = value & ((1 << self.m_size) - 1)

    def needs_update(self):
        """
           Function: needs_update

           Check if the abstract model contains different desired and mirrored values.

           If a desired field value has been modified in the abstraction class
           without actually updating the field in the DUT,
           the state of the DUT (more specifically what the abstraction class
           `thinks` the state of the DUT is) is outdated.
           This method returns TRUE
           if the state of the field in the DUT needs to be updated
           to match the desired value.
           The mirror values or actual content of DUT field are not modified.
           Use the `UVMReg.update` to actually update the DUT field.

        Returns:
        """
        return (self.m_mirrored != self.m_desired) or self.m_volatile

    #
    #   // Task: write
    #   //
    #   // Write the specified value in this field
    #   //
    #   // Write ~value~ in the DUT field that corresponds to this
    #   // abstraction class instance using the specified access
    #   // ~path~.
    #   // If the register containing this field is mapped in more
    #   //  than one address map,
    #   // an address ~map~ must be
    #   // specified if a physical access is used (front-door access).
    #   // If a back-door access path is used, the effect of writing
    #   // the field through a physical access is mimicked. For
    #   // example, read-only bits in the field will not be written.
    #   //
    #   // The mirrored value will be updated using the `UVMRegField.predict`
    #   // method.
    #   //
    #   // If a front-door access is used, and
    #   // if the field is the only field in a byte lane and
    #   // if the physical interface corresponding to the address map used
    #   // to access the field support byte-enabling,
    #   // then only the field is written.
    #   // Otherwise, the entire register containing the field is written,
    #   // and the mirrored values of the other fields in the same register
    #   // are used in a best-effort not to modify their value.
    #   //
    #   // If a backdoor access is used, a peek-modify-poke process is used.
    #   // in a best-effort not to modify the value of the other fields in the
    #   // register.
    #   //
    #   extern virtual task write (output uvm_status_e       status,
    #                              input  uvm_reg_data_t     value,
    #                              input  uvm_path_e         path = UVM_DEFAULT_PATH,
    #                              input  uvm_reg_map        map = null,
    #                              input  uvm_sequence_base  parent = null,
    #                              input  int                prior = -1,
    #                              input  uvm_object         extension = null,
    #                              input  string             fname = "",
    #                              input  int                lineno = 0)



    #   // Task: read
    #   //
    #   // Read the current value from this field
    #   //
    #   // Read and return ~value~ from the DUT field that corresponds to this
    #   // abstraction class instance using the specified access
    #   // ~path~.
    #   // If the register containing this field is mapped in more
    #   // than one address map, an address ~map~ must be
    #   // specified if a physical access is used (front-door access).
    #   // If a back-door access path is used, the effect of reading
    #   // the field through a physical access is mimicked. For
    #   // example, clear-on-read bits in the field will be set to zero.
    #   //
    #   // The mirrored value will be updated using the `UVMRegField.predict`
    #   // method.
    #   //
    #   // If a front-door access is used, and
    #   // if the field is the only field in a byte lane and
    #   // if the physical interface corresponding to the address map used
    #   // to access the field support byte-enabling,
    #   // then only the field is read.
    #   // Otherwise, the entire register containing the field is read,
    #   // and the mirrored values of the other fields in the same register
    #   // are updated.
    #   //
    #   // If a backdoor access is used, the entire containing register is peeked
    #   // and the mirrored value of the other fields in the register is updated.
    #   //
    #   extern virtual task read  (output uvm_status_e       status,
    #                              output uvm_reg_data_t     value,
    #                              input  uvm_path_e         path = UVM_DEFAULT_PATH,
    #                              input  uvm_reg_map        map = null,
    #                              input  uvm_sequence_base  parent = null,
    #                              input  int                prior = -1,
    #                              input  uvm_object         extension = null,
    #                              input  string             fname = "",
    #                              input  int                lineno = 0)
    #
    #

    #   // Task: poke
    #   //
    #   // Deposit the specified value in this field
    #   //
    #   // Deposit the value in the DUT field corresponding to this
    #   // abstraction class instance, as-is, using a back-door access.
    #   // A peek-modify-poke process is used
    #   // in a best-effort not to modify the value of the other fields in the
    #   // register.
    #   //
    #   // The mirrored value will be updated using the `UVMRegField.predict`
    #   // method.
    #   //
    #   extern virtual task poke  (output uvm_status_e       status,
    #                              input  uvm_reg_data_t     value,
    #                              input  string             kind = "",
    #                              input  uvm_sequence_base  parent = null,
    #                              input  uvm_object         extension = null,
    #                              input  string             fname = "",
    #                              input  int                lineno = 0)
    #
    #

    #   // Task: peek
    #   //
    #   // Read the current value from this field
    #   //
    #   // Sample the value in the DUT field corresponding to this
    #   // abstraction class instance using a back-door access.
    #   // The field value is sampled, not modified.
    #   //
    #   // Uses the HDL path for the design abstraction specified by ~kind~.
    #   //
    #   // The entire containing register is peeked
    #   // and the mirrored value of the other fields in the register
    #   // are updated using the `UVMRegField.predict` method.
    #   //
    #   //
    #   extern virtual task peek  (output uvm_status_e       status,
    #                              output uvm_reg_data_t     value,
    #                              input  string             kind = "",
    #                              input  uvm_sequence_base  parent = null,
    #                              input  uvm_object         extension = null,
    #                              input  string             fname = "",
    #                              input  int                lineno = 0)
    #
    #

    #   // Task: mirror
    #   //
    #   // Read the field and update/check its mirror value
    #   //
    #   // Read the field and optionally compared the readback value
    #   // with the current mirrored value if ~check~ is <UVself.m_check>.
    #   // The mirrored value will be updated using the <predict()>
    #   // method based on the readback value.
    #   //
    #   // The ~path~ argument specifies whether to mirror using
    #   // the  <UVM_FRONTDOOR> (<read>) or
    #   // <UVM_BACKDOOR> (<peek()>).
    #   //
    #   // If ~check~ is specified as <UVself.m_check>,
    #   // an error message is issued if the current mirrored value
    #   // does not match the readback value, unless <set_compare> was used
    #   // disable the check.
    #   //
    #   // If the containing register is mapped in multiple address maps and physical
    #   // access is used (front-door access), an address ~map~ must be specified.
    #   // For write-only fields, their content is mirrored and optionally
    #   // checked only if a UVM_BACKDOOR
    #   // access path is used to read the field.
    #   //
    #   extern virtual task mirror(output uvm_status_e      status,
    #                              input  uvm_check_e       check = UVM_NO_CHECK,
    #                              input  uvm_path_e        path = UVM_DEFAULT_PATH,
    #                              input  uvm_reg_map       map = null,
    #                              input  uvm_sequence_base parent = null,
    #                              input  int               prior = -1,
    #                              input  uvm_object        extension = null,
    #                              input  string            fname = "",
    #                              input  int               lineno = 0)
    #
    #

    def set_compare(self, check=UVM_CHECK):
        """
        Sets the compare policy during a mirror update.
        The field value is checked against its mirror only when both the
        `check` argument in `UVMRegBlock.mirror`, `UVMReg.mirror`,
        or `UVMRegField.mirror` and the compare policy for the
        field is `UVMRegField.m_check`.

        Args:
            check:
        """
        self.m_check = check



    def get_compare(self):
        """
        Returns the compare policy for this field.

        Returns:
        """
        return self.m_check

    #   // Function: is_indv_accessible
    #   //
    #   // Check if this field can be written individually, i.e. without
    #   // affecting other fields in the containing register.
    #   //
    #function bit uvm_reg_field::is_indv_accessible(uvm_path_e  path,
    #                                               uvm_reg_map local_map)
    #   if (path == UVM_BACKDOOR) begin
    #      `uvm_warning("RegModel",
    #         {"Individual BACKDOOR field access not available for field '",
    #         get_full_name(), "'. Accessing complete register instead."})
    #      return 0
    #   end
    #
    #   if (!self.m_individually_accessible) begin
    #         `uvm_warning("RegModel",
    #            {"Individual field access not available for field '",
    #            get_full_name(), "'. Accessing complete register instead."})
    #      return 0
    #   end
    #
    #   // Cannot access individual fields if the container register
    #   // has a user-defined front-door
    #   if (self.m_parent.get_frontdoor(local_map) != null) begin
    #      `uvm_warning("RegModel",
    #                   {"Individual field access not available for field '",
    #                    get_name(), "' because register '", self.m_parent.get_full_name(), "' has a user-defined front-door. Accessing complete register instead."})
    #      return 0
    #   end
    #
    #   begin
    #     uvm_reg_map system_map = local_map.get_root_map()
    #     uvm_reg_adapter adapter = system_map.get_adapter()
    #     if (adapter.supports_byte_enable)
    #       return 1
    #   end
    #
    #   begin
    #     int fld_idx
    #     int bus_width = local_map.get_n_bytes()
    #     uvm_reg_field fields[$]
    #     bit sole_field
    #
    #     self.m_parent.get_fields(fields)
    #
    #     if (fields.size() == 1) begin
    #        sole_field = 1
    #     end
    #     else begin
    #        int prev_lsb,this_lsb,next_lsb;
    #        int prev_sz,this_sz,next_sz;
    #        int bus_sz = bus_width*8
    #
    #        foreach (fields[i]) begin
    #           if (fields[i] == this) begin
    #              fld_idx = i
    #              break
    #           end
    #        end
    #
    #        this_lsb = fields[fld_idx].get_lsb_pos()
    #        this_sz  = fields[fld_idx].get_n_bits()
    #
    #        if (fld_idx>0) begin
    #          prev_lsb = fields[fld_idx-1].get_lsb_pos()
    #          prev_sz  = fields[fld_idx-1].get_n_bits()
    #        end
    #
    #        if (fld_idx < fields.size()-1) begin
    #          next_lsb = fields[fld_idx+1].get_lsb_pos()
    #          next_sz  = fields[fld_idx+1].get_n_bits()
    #        end
    #
    #        // if first field in register
    #        if (fld_idx == 0 &&
    #           ((next_lsb % bus_sz) == 0 ||
    #            (next_lsb - this_sz) > (next_lsb % bus_sz)))
    #           return 1
    #
    #        // if last field in register
    #        else if (fld_idx == (fields.size()-1) &&
    #            ((this_lsb % bus_sz) == 0 ||
    #             (this_lsb - (prev_lsb + prev_sz)) >= (this_lsb % bus_sz)))
    #           return 1
    #
    #        // if somewhere in between
    #        else begin
    #           if ((this_lsb % bus_sz) == 0) begin
    #              if ((next_lsb % bus_sz) == 0 ||
    #                  (next_lsb - (this_lsb + this_sz)) >= (next_lsb % bus_sz))
    #                 return 1
    #           end
    #           else begin
    #              if ( (next_lsb - (this_lsb + this_sz)) >= (next_lsb % bus_sz) &&
    #                  ((this_lsb - (prev_lsb + prev_sz)) >= (this_lsb % bus_sz)) )
    #                 return 1
    #           end
    #        end
    #     end
    #   end
    #
    #   `uvm_warning("RegModel",
    #       {"Target bus does not support byte enabling, and the field '",
    #       get_full_name(),"' is not the only field within the entire bus width. ",
    #       "Individual field access will not be available. ",
    #       "Accessing complete register instead."})
    #
    #   return 0
    #
    #endfunction


    def predict(self, value, be=-1, kind=UVM_PREDICT_DIRECT, path=UVM_FRONTDOOR,
            _map=None, fname="", lineno=0):
        """
        Update the mirrored and desired value for this field.

        Predict the mirror and desired value of the field based on the specified
        observed `value` on a bus using the specified address `map`.

        If `kind` is specified as `UVM_PREDICT_READ`, the value
        was observed in a read transaction on the specified address `map` or
        backdoor (if `path` is `UVM_BACKDOOR`).
        If `kind` is specified as `UVM_PREDICT_WRITE`, the value
        was observed in a write transaction on the specified address `map` or
        backdoor (if `path` is `UVM_BACKDOOR`).
        If `kind` is specified as `UVM_PREDICT_DIRECT`, the value
        was computed and is updated as-is, without regard to any access policy.
        For example, the mirrored value of a read-only field is modified
        by this method if `kind` is specified as `UVM_PREDICT_DIRECT`.

        This method does not allow an update of the mirror (or desired)
        when the register containing this field is busy executing
        a transaction because the results are unpredictable and
        indicative of a race condition in the testbench.

        Returns TRUE if the prediction was successful.

        #function bit uvm_reg_field::predict (uvm_reg_data_t    value,
                                            uvm_reg_byte_en_t be = -1,
                                            uvm_predict_e     kind = UVM_PREDICT_DIRECT,
                                            uvm_path_e        path = UVM_FRONTDOOR,
                                            uvm_reg_map       map = null,
                                            string            fname = "",
                                            int               lineno = 0)
        Args:
            value:
            be (int): Byte-enable
            kind:
            path: 
            _map (UVMRegMap):
            fname (str):
            lineno (int):
        Returns:
        """
        rw = UVMRegItem()
        rw.value[0] = value
        rw.path = path
        rw.map = _map
        rw.fname = fname
        rw.lineno = lineno
        self.do_predict(rw, kind, be)
        # predict = (rw.status == UVM_NOT_OK) ? 0 : 1
        if rw.status == UVM_NOT_OK:
            return 0
        return 1

    def _get_mask(self):
        return ((1 << self.m_size)-1)



    def XpredictX(self, cur_val, wr_val, _map):
        """
          /*local*/
          extern virtual function uvm_reg_data_t XpredictX (uvm_reg_data_t cur_val,
                                                            uvm_reg_data_t wr_val,
                                                            uvm_reg_map    map)
        Args:
            cur_val:
            wr_val:
            _map:
        Returns:
        """
        mask = self._get_mask()
        acc = self.get_access(_map)
        #   case (get_access(map))
        if acc == "RO":
            return cur_val
        elif acc == "RW":
            return wr_val
        elif acc == "RC":
            return cur_val
        elif acc == "RS":
            return cur_val
        elif acc == "WC":
            return 0
        elif acc == "WS":
            return mask
        elif acc == "WRC":
            return wr_val
        elif acc == "WRS":
            return wr_val
        elif acc == "WSRC":
            return mask
        elif acc == "WCRS":
            return 0
        elif acc == "W1C":
            return cur_val & (~wr_val)
        elif acc == "W1S":
            return cur_val | wr_val
        elif acc == "W1T":
            return cur_val ^ wr_val
        elif acc == "W0C":
            return cur_val & wr_val
        elif acc == "W0S":
            return cur_val | (~wr_val & mask)
        elif acc == "W0T":
            return cur_val ^ (~wr_val & mask)
        elif acc == "W1SRC":
            return cur_val | wr_val
        elif acc == "W1CRS":
            return cur_val & (~wr_val)
        elif acc == "W0SRC":
            return cur_val | (~wr_val & mask)
        elif acc == "W0CRS":
            return cur_val & wr_val
        elif acc == "WO":
            return wr_val
        elif acc == "WOC":
            return 0
        elif acc == "WOS":
            return mask
        elif acc == "W1":
            if (self.m_written):
                return cur_val
            else:
                return wr_val
        elif acc == "WO1":
            if (self.m_written):
                return cur_val
            else:
                return wr_val
        elif acc == "NOACCESS":
            return cur_val
        else:
            return wr_val

        uvm_fatal("RegModel", "uvm_reg_field::XpredictX(): Internal error")
        return 0
        #endfunction: XpredictX


    def XupdateX(self):
        """
        XupdateX
        Returns:
        """
        #   Figure out which value must be written to get the desired value
        #   given what we think is the current value in the hardware
        XupdateX = self.m_desired  # default action
        if self.m_access == "W1C":
            XupdateX = ~self.m_desired
        if self.m_access == "W1S":
            XupdateX = self.m_desired
        if self.m_access == "W1T":
            XupdateX = self.m_desired ^ self.m_mirrored
        if self.m_access == "W0C":
            XupdateX = self.m_desired
        if self.m_access == "W0S":
            XupdateX = ~self.m_desired
        if self.m_access == "W0T":
            XupdateX = ~(self.m_desired ^ self.m_mirrored)
        if self.m_access == "W1SRC":
            XupdateX = self.m_desired
        if self.m_access == "W1CRS":
            XupdateX = ~self.m_desired
        if self.m_access == "W0SRC":
            XupdateX = ~self.m_desired
        if self.m_access == "W0CRS":
            XupdateX = self.m_desired
        if self.m_access == "WO":
            XupdateX = self.m_desired
        if self.m_access == "WOC":
            XupdateX = self.m_desired
        if self.m_access == "WOS":
            XupdateX = self.m_desired
        if self.m_access == "W1":
            XupdateX = self.m_desired
        if self.m_access == "WO1":
            XupdateX = self.m_desired
        XupdateX = XupdateX & self._get_mask()
        return XupdateX


    #   /*local*/
    #   extern function bit Xcheck_accessX (input uvm_reg_item rw,
    #                                       output uvm_reg_map_info map_info,
    #                                       input string caller)

    #   extern virtual task do_write(uvm_reg_item rw)
    #   extern virtual task do_read(uvm_reg_item rw)

    def do_predict(self, rw, kind=UVM_PREDICT_DIRECT, be=-1):
        """
        do_predict
        Args:
            rw (UVMRegItem):
            kind:
            be (int): Byte-enable
        Raises:
        """
        field_val = rw.value[0] & self._get_mask()
        if rw.status != UVM_NOT_OK:
            rw.status = UVM_IS_OK

        # Assume that the entire field is enabled
        if (0x00000001 & be) == 0:
            raise Exception('DBG Early return field not enabled')

        self.m_fname = rw.fname
        self.m_lineno = rw.lineno
        #   case (kind)
        if kind == UVM_PREDICT_WRITE:
            #UVMRegFieldCbIter cbs = new(this)
            cbs = UVMRegFieldCbIter(self)
            if rw.path == UVM_FRONTDOOR or rw.path == UVM_PREDICT:
                field_val = self.XpredictX(self.m_mirrored, field_val, rw.map)
            self.m_written = True

            cb = cbs.first()
            while cb is not None:
                cb.post_predict(self, self.m_mirrored, field_val,
                        UVM_PREDICT_WRITE, rw.path, rw.map)
                cb = cbs.next()
            field_val &= self._get_mask()

        elif kind == UVM_PREDICT_READ:
            cbs = UVMRegFieldCbIter(self)
            if rw.path == UVM_FRONTDOOR or rw.path == UVM_PREDICT:
                acc = self.get_access(rw.map)
                if acc in ["RC", "WRC", "WSRC", "W1SRC", "W0SRC"]:
                    field_val = 0  # (clear)
                elif acc in ["RS", "WRS", "WCRS", "W1CRS", "W0CRS"]:
                    field_val = self._get_mask()  # all 1's (set)
                elif acc in ["WO", "WOC", "WOS", "WO1", "NOACCESS"]:
                    return
            # TODO callbacks

            cb = cbs.first()
            while cb is not None:
                cb.post_predict(self, self.m_mirrored, field_val,
                        UVM_PREDICT_READ, rw.path, rw.map)
                cb = cbs.next()
            field_val &= self._get_mask()

        elif UVM_PREDICT_DIRECT:
            if self.m_parent.is_busy():
                uvm_report_warning("RegModel", ("Trying to predict value of field '"
                   + self.get_name() + "' while register '" + self.m_parent.get_full_name()
                   + "' is being accessed"))
                rw.status = UVM_NOT_OK
        #   endcase

        # update the mirror with predicted value
        self.m_mirrored = field_val
        self.m_desired  = field_val
        self.value = field_val


    def pre_randomize(self):
        #   // Update the only publicly known property with the current
        #   // desired value so it can be used as a state variable should
        #   // the rand_mode of the field be turned off.
        self.value = self.m_desired


    def post_randomize(self):
        self.m_desired = self.value


    #   //-----------------
    #   // Group: Callbacks
    #   //-----------------
    #   `uvm_register_cb(uvm_reg_field, uvm_reg_cbs)

    async def pre_write(self, rw):
        """
           Called before field write.

           If the specified data value, access `path` or address `map` are modified,
           the updated data value, access path or address map will be used
           to perform the register operation.
           If the `status` is modified to anything other than `UVM_IS_OK`,
           the operation is aborted.

           The field callback methods are invoked after the callback methods
           on the containing register.
           The registered callback methods are invoked after the invocation
           of this method.

        Args:
            rw (UVMRegItem): Reg item associated with write.
        """
        #await uvm_zero_delay()
        pass

    async def post_write(self, rw):
        """
           Called after field write.

           If the specified `status` is modified,
           the updated status will be
           returned by the register operation.

           The field callback methods are invoked after the callback methods
           on the containing register.
           The registered callback methods are invoked before the invocation
           of this method.

        Args:
            rw (UVMRegItem): Reg item associated with write.
        """
        #await uvm_zero_delay()
        pass

    async def pre_read(self, rw):
        """
           Called before field read.

           If the access `path` or address `map` in the `rw` argument are modified,
           the updated access path or address map will be used to perform
           the register operation.
           If the `status` is modified to anything other than `UVM_IS_OK`,
           the operation is aborted.

           The field callback methods are invoked after the callback methods
           on the containing register.
           The registered callback methods are invoked after the invocation
           of this method.

        Args:
            rw (UVMRegItem): Reg item associated with read.
        """
        pass

    async def post_read(self, rw):
        """
           Task: post_read

           Called after field read.

           If the specified readback data or`status` in the `rw` argument is
           modified, the updated readback data or status will be
           returned by the register operation.

           The field callback methods are invoked after the callback methods
           on the containing register.
           The registered callback methods are invoked before the invocation
           of this method.

        Args:
            rw (UVMRegItem): Reg item associated with read.
        """
        pass

    #   extern virtual function void do_print (uvm_printer printer)

    def convert2string(self):
        """
        convert2string

        Returns:
            str:
        """
        fmt = ""
        res_str = ""
        t_str = ""
        with_debug_info = False
        prefix = ""
        reg_ = self.get_register()
        convert2string = ""

        fmt = "{}'h%{}h".format(self.get_n_bits(), (self.get_n_bits()-1)/4 + 1)

        mirrored = ""
        if self.m_desired != self.m_mirrored:
            mirrored = (" (Mirror: " + fmt + ")").format(self.m_mirrored)

        convert2string = ("{} {} {}[{}:{}]=" + fmt + "{}").format(prefix,
            self.get_access(),
            reg_.get_name(),
            self.get_lsb_pos() + self.get_n_bits() - 1,
            self.get_lsb_pos(), self.m_desired,
            mirrored)

        if self.m_read_in_progress == 1:
            if self.m_fname != "" and self.m_lineno != 0:
                res_str = " from {}:{}".format(self.m_fname, self.m_lineno)
            convert2string = convert2string + "\n" + "currently being read" + res_str
        if self.m_write_in_progress == 1:
            if self.m_fname != "" and self.m_lineno != 0:
                res_str += " from {}:{}".format(self.m_fname, self.m_lineno)
            convert2string = convert2string + "\n" + res_str + "currently being written"
        return convert2string


    #   extern virtual function uvm_object clone()
    #   extern virtual function void do_copy   (uvm_object rhs)
    #   extern virtual function bit  do_compare (uvm_object  rhs,
    #                                            uvm_comparer comparer)
    #   extern virtual function void do_pack (uvm_packer packer)
    #   extern virtual function void do_unpack (uvm_packer packer)


uvm_object_utils(UVMRegField)

#//------------------------------------------------------------------------------
#// IMPLEMENTATION
#//------------------------------------------------------------------------------
#
#// m_predefined_policies
#
#typedef class uvm_reg_map_info
#
#// Xcheck_accessX
#
#function bit uvm_reg_field::Xcheck_accessX(input uvm_reg_item rw,
#                                           output uvm_reg_map_info map_info,
#                                           input string caller)
#
#
#   if (rw.path == UVM_DEFAULT_PATH) begin
#     uvm_reg_block blk = self.m_parent.get_block()
#     rw.path = blk.get_default_path()
#   end
#
#   if (rw.path == UVM_BACKDOOR) begin
#      if (self.m_parent.get_backdoor() == null && !self.m_parent.has_hdl_path()) begin
#         `uvm_warning("RegModel",
#            {"No backdoor access available for field '",get_full_name(),
#            "' . Using frontdoor instead."})
#         rw.path = UVM_FRONTDOOR
#      end
#      else
#        rw.map = uvm_reg_map::backdoor()
#   end
#
#   if (rw.path != UVM_BACKDOOR) begin
#
#     rw.local_map = self.m_parent.get_local_map(rw.map,caller)
#
#     if (rw.local_map == null) begin
#        `uvm_error(get_type_name(),
#           {"No transactor available to physically access memory from map '",
#            rw.map.get_full_name(),"'"})
#        rw.status = UVM_NOT_OK
#        return 0
#     end
#
#     map_info = rw.local_map.get_reg_map_info(self.m_parent)
#
#     if (map_info.frontdoor == null && map_info.unmapped) begin
#        `uvm_error("RegModel", {"Field '",get_full_name(),
#                   "' in register that is unmapped in map '",
#                   rw.map.get_full_name(),
#                   "' and does not have a user-defined frontdoor"})
#        rw.status = UVM_NOT_OK
#        return 0
#     end
#
#     if (rw.map == null)
#       rw.map = rw.local_map
#   end
#
#   return 1
#endfunction
#
#
#// write
#
#task uvm_reg_field::write(output uvm_status_e       status,
#                          input  uvm_reg_data_t     value,
#                          input  uvm_path_e         path = UVM_DEFAULT_PATH,
#                          input  uvm_reg_map        map = null,
#                          input  uvm_sequence_base  parent = null,
#                          input  int                prior = -1,
#                          input  uvm_object         extension = null,
#                          input  string             fname = "",
#                          input  int                lineno = 0)
#
#   uvm_reg_item rw
#   rw = uvm_reg_item::type_id::create("field_write_item",,get_full_name())
#   rw.element      = this
#   rw.element_kind = UVM_FIELD
#   rw.kind         = UVM_WRITE
#   rw.value[0]     = value
#   rw.path         = path
#   rw.map          = map
#   rw.parent       = parent
#   rw.prior        = prior
#   rw.extension    = extension
#   rw.fname        = fname
#   rw.lineno       = lineno
#
#   do_write(rw)
#
#   status = rw.status
#
#endtask
#
#
#// do_write
#
#task uvm_reg_field::do_write(uvm_reg_item rw)
#
#   uvm_reg_data_t   value_adjust
#   uvm_reg_map_info map_info
#   uvm_reg_field    fields[$]
#   bit bad_side_effect
#
#   self.m_parent.XatomicX(1)
#   self.m_fname  = rw.fname
#   self.m_lineno = rw.lineno
#
#   if (!Xcheck_accessX(rw,map_info,"write()"))
#     return
#
#   self.m_write_in_progress = 1'b1
#
#   if (rw.value[0] >> self.m_size) begin
#      `uvm_warning("RegModel", {"uvm_reg_field::write(): Value greater than field '",
#                          get_full_name(),"'"})
#      rw.value[0] &= ((1<<self.m_size)-1)
#   end
#
#   // Get values to write to the other fields in register
#   self.m_parent.get_fields(fields)
#   foreach (fields[i]) begin
#
#      if (fields[i] == this) begin
#         value_adjust |= rw.value[0] << self.m_lsb
#         continue
#      end
#
#      // It depends on what kind of bits they are made of...
#      case (fields[i].get_access(rw.local_map))
#        // These...
#        "RO", "RC", "RS", "W1C", "W1S", "W1T", "W1SRC", "W1CRC":
#          // Use all 0's
#          value_adjust |= 0
#
#        // These...
#        "W0C", "W0S", "W0T", "W0SRC", "W0CRS":
#          // Use all 1's
#          value_adjust |= ((1<<fields[i].get_n_bits())-1) << fields[i].get_lsb_pos()
#
#        // These might have side effects! Bad!
#        "WC", "WS", "WCRS", "WSRC", "WOC", "WOS":
#           bad_side_effect = 1
#
#        default:
#           value_adjust |= fields[i].m_mirrored << fields[i].get_lsb_pos()
#
#      endcase
#   end
#
#`ifdef UVM_REG_NO_INDIVIDUAL_FIELD_ACCESS
#   rw.element_kind = UVM_REG
#   rw.element = self.m_parent
#   rw.value[0] = value_adjust
#   self.m_parent.do_write(rw);
#`else
#
#   if (!is_indv_accessible(rw.path,rw.local_map)) begin
#      rw.element_kind = UVM_REG
#      rw.element = self.m_parent
#      rw.value[0] = value_adjust
#      self.m_parent.do_write(rw)
#
#      if (bad_side_effect) begin
#         `uvm_warning("RegModel", $sformatf("Writing field \"%s\" will cause unintended side effects in adjoining Write-to-Clear or Write-to-Set fields in the same register", this.get_full_name()))
#      end
#   end
#   else begin
#
#     uvm_reg_map system_map = rw.local_map.get_root_map()
#     UVMRegFieldCbIter cbs = new(this)
#
#     self.m_parent.Xset_busyX(1)
#
#     rw.status = UVM_IS_OK
#
#     pre_write(rw)
#     for (uvm_reg_cbs cb=cbs.first(); cb!=null; cb=cbs.next())
#        cb.pre_write(rw)
#
#     if (rw.status != UVM_IS_OK) begin
#        self.m_write_in_progress = 1'b0
#        self.m_parent.Xset_busyX(0)
#        self.m_parent.XatomicX(0)
#
#        return
#     end
#
#     rw.local_map.do_write(rw)
#
#     if (system_map.get_auto_predict())
#        // ToDo: Call parent.XsampleX()
#        do_predict(rw, UVM_PREDICT_WRITE)
#
#     post_write(rw)
#     for (uvm_reg_cbs cb=cbs.first(); cb!=null; cb=cbs.next())
#        cb.post_write(rw)
#
#     self.m_parent.Xset_busyX(0)
#
#   end
#
#`endif
#
#   self.m_write_in_progress = 1'b0
#   self.m_parent.XatomicX(0)
#
#endtask: do_write
#
#
#// read
#
#task uvm_reg_field::read(output uvm_status_e       status,
#                         output uvm_reg_data_t     value,
#                         input  uvm_path_e         path = UVM_DEFAULT_PATH,
#                         input  uvm_reg_map        map = null,
#                         input  uvm_sequence_base  parent = null,
#                         input  int                prior = -1,
#                         input  uvm_object         extension = null,
#                         input  string             fname = "",
#                         input  int                lineno = 0)
#
#   uvm_reg_item rw
#   rw = uvm_reg_item::type_id::create("field_read_item",,get_full_name())
#   rw.element      = this
#   rw.element_kind = UVM_FIELD
#   rw.kind         = UVM_READ
#   rw.value[0]     = 0
#   rw.path         = path
#   rw.map          = map
#   rw.parent       = parent
#   rw.prior        = prior
#   rw.extension    = extension
#   rw.fname        = fname
#   rw.lineno       = lineno
#
#   do_read(rw)
#
#   value = rw.value[0]
#   status = rw.status
#
#endtask: read
#
#
#// do_read
#
#task uvm_reg_field::do_read(uvm_reg_item rw)
#
#   uvm_reg_map_info map_info
#   bit bad_side_effect
#
#   self.m_parent.XatomicX(1)
#   self.m_fname  = rw.fname
#   self.m_lineno = rw.lineno
#   self.m_read_in_progress = 1'b1
#
#   if (!Xcheck_accessX(rw,map_info,"read()"))
#     return
#
#`ifdef UVM_REG_NO_INDIVIDUAL_FIELD_ACCESS
#   rw.element_kind = UVM_REG
#   rw.element = self.m_parent
#   self.m_parent.do_read(rw)
#   rw.value[0] = (rw.value[0] >> self.m_lsb) & ((1<<self.m_size))-1
#   bad_side_effect = 1
#`else
#
#   if (!is_indv_accessible(rw.path,rw.local_map)) begin
#      rw.element_kind = UVM_REG
#      rw.element = self.m_parent
#      bad_side_effect = 1
#      self.m_parent.do_read(rw)
#      rw.value[0] = (rw.value[0] >> self.m_lsb) & ((1<<self.m_size))-1
#   end
#   else begin
#
#     uvm_reg_map system_map = rw.local_map.get_root_map()
#     UVMRegFieldCbIter cbs = new(this)
#
#     self.m_parent.Xset_busyX(1)
#
#     rw.status = UVM_IS_OK
#
#     pre_read(rw)
#     for (uvm_reg_cbs cb = cbs.first(); cb != null; cb = cbs.next())
#        cb.pre_read(rw)
#
#     if (rw.status != UVM_IS_OK) begin
#        self.m_read_in_progress = 1'b0
#        self.m_parent.Xset_busyX(0)
#        self.m_parent.XatomicX(0)
#
#        return
#     end
#
#     rw.local_map.do_read(rw)
#
#
#     if (system_map.get_auto_predict())
#        // ToDo: Call parent.XsampleX()
#        do_predict(rw, UVM_PREDICT_READ)
#
#     post_read(rw)
#     for (uvm_reg_cbs cb=cbs.first(); cb!=null; cb=cbs.next())
#        cb.post_read(rw)
#
#     self.m_parent.Xset_busyX(0)
#
#   end
#
#`endif
#
#   self.m_read_in_progress = 1'b0
#   self.m_parent.XatomicX(0)
#
#   if (bad_side_effect) begin
#      uvm_reg_field fields[$]
#      self.m_parent.get_fields(fields)
#      foreach (fields[i]) begin
#         string mode
#         if (fields[i] == this)
#            continue
#         mode = fields[i].get_access()
#         if (mode == "RC" ||
#             mode == "RS" ||
#             mode == "WRC" ||
#             mode == "WRS" ||
#             mode == "WSRC" ||
#             mode == "WCRS" ||
#             mode == "W1SRC" ||
#             mode == "W1CRS" ||
#             mode == "W0SRC" ||
#             mode == "W0CRS") begin
#            `uvm_warning("RegModel", {"Reading field '",get_full_name(),
#                "' will cause unintended side effects in adjoining ",
#                "Read-to-Clear or Read-to-Set fields in the same register"})
#         end
#      end
#   end
#
#endtask: do_read
#
#
#// is_indv_accessible
#
#
#
#// poke
#
#task uvm_reg_field::poke(output uvm_status_e      status,
#                         input  uvm_reg_data_t    value,
#                         input  string            kind = "",
#                         input  uvm_sequence_base parent = null,
#                         input  uvm_object        extension = null,
#                         input  string            fname = "",
#                         input  int               lineno = 0)
#   uvm_reg_data_t  tmp
#
#   self.m_fname = fname
#   self.m_lineno = lineno
#
#   if (value >> self.m_size) begin
#      `uvm_warning("RegModel",
#         {"uvm_reg_field::poke(): Value exceeds size of field '",
#          get_name(),"'"})
#      value &= value & ((1<<self.m_size)-1)
#   end
#
#
#   self.m_parent.XatomicX(1)
#   self.m_parent.m_is_locked_by_field = 1'b1
#
#   tmp = 0
#
#   // What is the current values of the other fields???
#   self.m_parent.peek(status, tmp, kind, parent, extension, fname, lineno)
#
#   if (status == UVM_NOT_OK) begin
#      `uvm_error("RegModel", {"uvm_reg_field::poke(): Peek of register '",
#         self.m_parent.get_full_name(),"' returned status ",status.name()})
#      self.m_parent.XatomicX(0)
#      self.m_parent.m_is_locked_by_field = 1'b0
#      return
#   end
#
#   // Force the value for this field then poke the resulting value
#   tmp &= ~(((1<<self.m_size)-1) << self.m_lsb)
#   tmp |= value << self.m_lsb
#   self.m_parent.poke(status, tmp, kind, parent, extension, fname, lineno)
#
#   self.m_parent.XatomicX(0)
#   self.m_parent.m_is_locked_by_field = 1'b0
#endtask: poke
#
#
#// peek
#
#task uvm_reg_field::peek(output uvm_status_e      status,
#                         output uvm_reg_data_t    value,
#                         input  string            kind = "",
#                         input  uvm_sequence_base parent = null,
#                         input  uvm_object        extension = null,
#                         input  string            fname = "",
#                         input  int               lineno = 0)
#   uvm_reg_data_t  reg_value
#
#   self.m_fname = fname
#   self.m_lineno = lineno
#
#   self.m_parent.peek(status, reg_value, kind, parent, extension, fname, lineno)
#   value = (reg_value >> self.m_lsb) & ((1<<self.m_size))-1
#
#endtask: peek
#
#
#// mirror
#
#task uvm_reg_field::mirror(output uvm_status_e      status,
#                           input  uvm_check_e       check = UVM_NO_CHECK,
#                           input  uvm_path_e        path = UVM_DEFAULT_PATH,
#                           input  uvm_reg_map       map = null,
#                           input  uvm_sequence_base parent = null,
#                           input  int               prior = -1,
#                           input  uvm_object        extension = null,
#                           input  string            fname = "",
#                           input  int               lineno = 0)
#   self.m_fname = fname
#   self.m_lineno = lineno
#   self.m_parent.mirror(status, check, path, map, parent, prior, extension,
#                      fname, lineno)
#endtask: mirror
#
#
#
#// do_print
#
#function void uvm_reg_field::do_print (uvm_printer printer)
#  printer.print_generic(get_name(), get_type_name(), -1, convert2string())
#endfunction
#
#
#
#
#// clone
#
#function uvm_object uvm_reg_field::clone()
#  `uvm_fatal("RegModel","RegModel field cannot be cloned")
#  return null
#endfunction
#
#// do_copy
#
#function void uvm_reg_field::do_copy(uvm_object rhs)
#  `uvm_warning("RegModel","RegModel field copy not yet implemented")
#  // just a set(rhs.get()) ?
#endfunction
#
#
#// do_compare
#
#function bit uvm_reg_field::do_compare (uvm_object  rhs,
#                                        uvm_comparer comparer)
#  `uvm_warning("RegModel","RegModel field compare not yet implemented")
#  // just a return (get() == rhs.get()) ?
#  return 0
#endfunction
#
#
#// do_pack
#
#function void uvm_reg_field::do_pack (uvm_packer packer)
#  `uvm_warning("RegModel","RegModel field cannot be packed")
#endfunction
#
#
#// do_unpack
#
#function void uvm_reg_field::do_unpack (uvm_packer packer)
#  `uvm_warning("RegModel","RegModel field cannot be unpacked")
#endfunction
#




import unittest
from uvm import (UVM_IMP_COMMON, UVM_TLM_NONBLOCKING_PUT_MASK)


class TestUVMTLMImps(unittest.TestCase):

    def test_uvm_imp_common(self):
        from uvm.base.uvm_port_base import UVMPortBase
        from uvm.base.uvm_component import UVMComponent

        class MyTest():
            pass
        MyTest = UVM_IMP_COMMON(MyTest, UVM_TLM_NONBLOCKING_PUT_MASK, 'MyTest')
        obj = MyTest('MyPort', UVMComponent('xxx', None))
        self.assertEqual(obj.get_type_name(), 'MyTest')
        self.assertEqual(obj.is_imp(), True)
        self.assertEqual(isinstance(obj, UVMPortBase), True)


if __name__ == '__main__':
    unittest.main()



#------------------------------------------------------------------------------
#   Copyright 2007-2010 Mentor Graphics Corporation
#   Copyright 2007-2009 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------

from .uvm_callback import (UVMCallback, UVMCallbacks, UVMCallbackIter,
    UVMCallbacksBase)
from .uvm_object_globals import (UVM_ERROR, UVM_FATAL, UVM_NONE, UVM_WARNING,
    UVM_LOG, UVM_RM_RECORD, UVM_LOW, UVM_INFO, UVM_NO_ACTION, UVM_DISPLAY)
from .sv import sv
from ..macros import uvm_info_context
from ..uvm_macros import UVM_STRING_QUEUE_STREAMING_PACK
from .uvm_report_message import UVMReportMessage
from .uvm_globals import uvm_report_enabled


class sev_id_struct:
    def __init__(self):
        self.sev_specified = False
        self.id_specified = False
        self.sev = 0
        self.id = ""
        self.is_on = False


# tpoikela: Not typedef like in SV-UVM, but special class with
# CB=UVMReportCatcher
class UVMReportCb(UVMCallbacks):

    def __init__(self, name='uvm_report_cb'):
        super().__init__(name, T=None, CB=UVMReportCatcher)

    @classmethod
    def get_first(cls, itr, obj):
        return super().get_first(itr, obj, UVMReportCatcher)


UVMReportCbIter = UVMCallbackIter




UNKNOWN_ACTION = 0
THROW = 1
CAUGHT = 2


class UVMReportCatcher(UVMCallback):
    """
    The UVMReportCatcher is used to catch messages issued by the uvm report
    server. Catchers are
    UVMCallbacks objects,
    so all facilities in the `UVMCallback` and `UVMCallbacks`
    classes are available for registering catchers and controlling catcher
    state.

    The `UVMCallbacks` class is
    aliased to `UVMReportCb` to make it easier to use.
    Multiple report catchers can be
    registered with a report object. The catchers can be registered as default
    catchers which catch all reports on all `UVMReportObject` reporters,
    or catchers can be attached to specific report objects (i.e. components).

    User extensions of `UVMReportCatcher` must implement the <catch> method in
    which the action to be taken on catching the report is specified. The catch
    method can return ~CAUGHT~, in which case further processing of the report is
    immediately stopped, or return ~THROW~ in which case the (possibly modified) report
    is passed on to other registered catchers. The catchers are processed in the order
    in which they are registered.

    On catching a report, the <catch> method can modify the severity, id, action,
    verbosity or the report string itself before the report is finally issued by
    the report server. The report can be immediately issued from within the catcher
    class by calling the <issue> method.

    The catcher maintains a count of all reports with FATAL,ERROR or WARNING severity
    and a count of all reports with FATAL, ERROR or WARNING severity whose severity
    was lowered. These statistics are reported in the summary of the <uvm_report_server>.

    This example shows the basic concept of creating a report catching
    callback and attaching it to all messages that get emitted::

        class my_error_demoter(UVMReportCatcher):
            def __init__(self, name="my_error_demoter"):
                super().__init__(name)

            # This example demotes "MY_ID" errors to an info message
            def catch(self):
                if self.get_severity() == UVM_ERROR and self.get_id() == "MY_ID":
                    self.set_severity(UVM_INFO)
                return THROW

        @cocotb.test()
        async initial_begin(dut):
            demoter = my_error_demoter()
            # Catchers are callbacks on report objects (components are report
            # objects, so catchers can be attached to components).

            # To affect all reporters, use ~null~ for the object
            UVMReportCb.add(None, demoter)

            # To affect some specific object use the specific reporter
            UVMReportCb.add(mytest.myenv.myagent.mydriver, demoter)

            # To affect some set of components (any "*driver" under mytest.myenv)
            # using the component name
            UVMReportCb.add_by_name("*driver", demoter, mytest.myenv)

    """


    #`uvm_register_cb(uvm_report_object,uvm_report_catcher)
    m_modified_report_message = None  # type: UVMReportMessage
    m_orig_report_message = None  # type: UVMReportMessage
    m_set_action_called = False

    # Counts for the demoteds and caughts
    m_demoted_fatal = 0
    m_demoted_error = 0
    m_demoted_warning = 0
    m_caught_fatal = 0
    m_caught_error = 0
    m_caught_warning = 0

    #// Flag counts
    DO_NOT_CATCH      = 1
    DO_NOT_MODIFY     = 2
    m_debug_flags = 0

    do_report = False

    #// Function: new
    #//
    #// Create a new report catcher. The name argument is optional, but
    #// should generally be provided to aid in debugging.
    def __init__(self, name="uvm_report_catcher"):
        super().__init__(name)
        UVMReportCatcher.do_report = True

    #// Group: Current Message State
    #
    #// Function: get_client
    #//
    #// Returns the <uvm_report_object> that has generated the message that
    #// is currently being processed.
    def get_client(self):
        return UVMReportCatcher.m_modified_report_message.get_report_object()

    #// Function: get_severity
    #//
    #// Returns the <uvm_severity> of the message that is currently being
    #// processed. If the severity was modified by a previously executed
    #// catcher object (which re-threw the message), then the returned
    #// severity is the modified value.
    def get_severity(self):
        return UVMReportCatcher.m_modified_report_message.get_severity()

    #// Function: get_context
    #//
    #// Returns the context name of the message that is currently being
    #// processed. This is typically the full hierarchical name of the component
    #// that issued the message. However, if user-defined context is set from
    #// a uvm_report_message, the user-defined context will be returned.
    def get_context(self):
        context_str = self.m_modified_report_message.get_context()
        if (context_str == ""):
            rh = self.m_modified_report_message.get_report_handler()
            context_str = rh.get_full_name()
        return context_str


    #// Function: get_verbosity
    #//
    #// Returns the verbosity of the message that is currently being
    #// processed. If the verbosity was modified by a previously executed
    #// catcher (which re-threw the message), then the returned
    #// verbosity is the modified value.
    def get_verbosity(self):
        return self.m_modified_report_message.get_verbosity()


    #// Function: get_id
    #//
    #// Returns the string id of the message that is currently being
    #// processed. If the id was modified by a previously executed
    #// catcher (which re-threw the message), then the returned
    #// id is the modified value.
    def get_id(self):
        return self.m_modified_report_message.get_id()


    #// Function: get_message
    #//
    #// Returns the string message of the message that is currently being
    #// processed. If the message was modified by a previously executed
    #// catcher (which re-threw the message), then the returned
    #// message is the modified value.
    def get_message(self):
        return self.m_modified_report_message.get_message()


    #// Function: get_action
    #//
    #// Returns the <uvm_action> of the message that is currently being
    #// processed. If the action was modified by a previously executed
    #// catcher (which re-threw the message), then the returned
    #// action is the modified value.
    def get_action(self):
        return self.m_modified_report_message.get_action()


    #// Function: get_fname
    #//
    #// Returns the file name of the message.
    def get_fname(self):
        return self.m_modified_report_message.get_filename()


    #// Function: get_line
    #//
    #// Returns the line number of the message.
    def get_line(self):
        return self.m_modified_report_message.get_line()


    #// Function: get_element_container
    #//
    #// Returns the element container of the message.
    def get_element_container(self):
        return self.m_modified_report_message.get_element_container()


    #// Group: Change Message State
    #
    #// Function: set_severity
    #//
    #// Change the severity of the message to ~severity~. Any other
    #// report catchers will see the modified value.
    def set_severity(self, severity):
        self.m_modified_report_message.set_severity(severity)


    #// Function: set_verbosity
    #//
    #// Change the verbosity of the message to ~verbosity~. Any other
    #// report catchers will see the modified value.
    #
    def set_verbosity(self, verbosity):
        self.m_modified_report_message.set_verbosity(verbosity)


    #// Function: set_id
    #//
    #// Change the id of the message to ~id~. Any other
    #// report catchers will see the modified value.
    #
    def set_id(self, id):
        self.m_modified_report_message.set_id(id)


    #// Function: set_message
    #//
    #// Change the text of the message to ~message~. Any other
    #// report catchers will see the modified value.
    def set_message(self, message):
        self.m_modified_report_message.set_message(message)


    #// Function: set_action
    #//
    #// Change the action of the message to ~action~. Any other
    #// report catchers will see the modified value.
    #
    def set_action(self, action):
        self.m_modified_report_message.set_action(action)
        self.m_set_action_called = 1

    #// Function: set_context
    #//
    #// Change the context of the message to ~context_str~. Any other
    #// report catchers will see the modified value.
    def set_context(self, context_str):
        self.m_modified_report_message.set_context(context_str)


    #// Function: add_int
    #//
    #// Add an integral type of the name ~name~ and value ~value~ to
    #// the message.  The required ~size~ field indicates the size of ~value~.
    #// The required ~radix~ field determines how to display and
    #// record the field. Any other report catchers will see the newly
    #// added element.
    #//
    def add_int(self, name, value, size, radix, action=(UVM_LOG | UVM_RM_RECORD)):
        self.m_modified_report_message.add_int(name, value, size, radix, action)


    #// Function: add_string
    #//
    #// Adds a string of the name ~name~ and value ~value~ to the
    #// message. Any other report catchers will see the newly
    #// added element.
    #//
    def add_string(self, name, value, action=(UVM_LOG | UVM_RM_RECORD)):
        self.m_modified_report_message.add_string(name, value, action)


    #// Function: add_object
    #//
    #// Adds a uvm_object of the name ~name~ and reference ~obj~ to
    #// the message. Any other report catchers will see the newly
    #// added element.
    #//
    def add_object(self, name, obj, action=(UVM_LOG | UVM_RM_RECORD)):
        self.m_modified_report_message.add_object(name, obj, action)


    #// Group: Debug

    cb_iter = None  # static variable for get_report_catcher iterator

    #// Function: get_report_catcher
    #//
    #// Returns the first report catcher that has ~name~.
    @classmethod
    def get_report_catcher(cls, name):
        cls.cb_iter = UVMReportCbIter(None)
        get_report_catcher = cls.cb_iter.first()
        while get_report_catcher is not None:
            if get_report_catcher.get_name() == name:
                return get_report_catcher
            get_report_catcher = cls.cb_iter.next()
        return None


    #// Function: print_catcher
    #//
    #// Prints information about all of the report catchers that are
    #// registered. For finer grained detail, the <uvm_callbacks #(T,CB)::display>
    #// method can be used by calling uvm_report_cb::display(<uvm_report_object>).
    @classmethod
    def print_catcher(cls, file=0):
        # msg = ""
        enabled = ""
        catcher = None  # UVMReportCatcher
        cls.cb_iter = UVMReportCbIter(None)
        q = []

        q.append("-------------UVM REPORT CATCHERS----------------------------\n")

        catcher = cls.cb_iter.first()
        while catcher is not None:
            if catcher.callback_mode():
                enabled = "ON"
            else:
                enabled = "OFF"

            q.append(sv.sformatf("%20s : %s\n", catcher.get_name(),enabled))
            catcher = cls.cb_iter.next()

        q.append("--------------------------------------------------------------\n")
        from .uvm_root import uvm_top
        uvm_info_context("UVM/REPORT/CATCHER", UVM_STRING_QUEUE_STREAMING_PACK(q), UVM_LOW, uvm_top)


    #// Function: debug_report_catcher
    #//
    #// Turn on report catching debug information. ~what~ is a bitwise AND of
    #// * DO_NOT_CATCH  -- forces catch to be ignored so that all catchers see the
    #//   the reports.
    #// * DO_NOT_MODIFY -- forces the message to remain unchanged
    @classmethod
    def debug_report_catcher(cls, what=0):
        cls.m_debug_flags = what

    #// Group: Callback Interface

    #// Function: catch
    #//
    #// This is the method that is called for each registered report catcher.
    #// There are no arguments to this function. The <Current Message State>
    #// interface methods can be used to access information about the
    #// current message being processed.
    def catch(self):
        raise Exception("catch() must be implemented in user-class")


    #// Group: Reporting

    # // Function: uvm_report_fatal
    # //
    # // Issues a fatal message using the current message's report object.
    # // This message will bypass any message catching callbacks.
    #
    def uvm_report_fatal(self, id, message, verbosity, fname="", line=0,
            context_name="", report_enabled_checked=0):

        self.uvm_report(UVM_FATAL, id, message, UVM_NONE, fname, line,
            context_name, report_enabled_checked)


    # // Function: uvm_report_error
    # //
    # // Issues an error message using the current message's report object.
    # // This message will bypass any message catching callbacks.
    #
    def uvm_report_error(self, id, message, verbosity, fname="", line=0,
            context_name="", report_enabled_checked=0):
        self.uvm_report(UVM_ERROR, id, message, UVM_NONE, fname, line,
            context_name, report_enabled_checked)

    # // Function: uvm_report_warning
    # //
    # // Issues a warning message using the current message's report object.
    # // This message will bypass any message catching callbacks.
    #
    def uvm_report_warning(self, id, message, verbosity, fname="", line=0,
            context_name="", report_enabled_checked=0):
        self.uvm_report(UVM_WARNING, id, message, UVM_NONE, fname, line,
            context_name, report_enabled_checked)


    # // Function: uvm_report_info
    # //
    # // Issues a info message using the current message's report object.
    # // This message will bypass any message catching callbacks.
    #
    def uvm_report_info(self, id, message, verbosity, fname="", line=0,
            context_name="", report_enabled_checked=0):
        self.uvm_report(UVM_INFO, id, message, verbosity, fname, line,
            context_name, report_enabled_checked)


    # // Function: uvm_report
    # //
    # // Issues a message using the current message's report object.
    # // This message will bypass any message catching callbacks.
    #
    def uvm_report(self, severity,
        id,
        message,
        verbosity,
        fname="",
        line=0,
        context_name="",
        report_enabled_checked=0):

        l_report_message = None
        if report_enabled_checked == 0:
            if not uvm_report_enabled(verbosity, severity, id):
                return

        l_report_message = UVMReportMessage.new_report_message()
        l_report_message.set_report_message(severity, id, message,
                            verbosity, fname, line, context_name)
        self.uvm_process_report_message(l_report_message)


    def uvm_process_report_message(self, msg: UVMReportMessage):
        ro = UVMReportCatcher.m_modified_report_message.get_report_object()
        a = ro.get_report_action(msg.get_severity(), msg.get_id())

        if a:
            composed_message = ""
            rs = UVMReportCatcher.m_modified_report_message.get_report_server()

            msg.set_report_object(ro)
            msg.set_report_handler(UVMReportCatcher.m_modified_report_message.get_report_handler())
            msg.set_report_server(rs)
            msg.set_file(ro.get_report_file_handle(msg.get_severity(), msg.get_id()))
            msg.set_action(a)

            # no need to compose when neither UVM_DISPLAY nor UVM_LOG is set
            if a & (UVM_LOG | UVM_DISPLAY):
                composed_message = rs.compose_report_message(msg)
            rs.execute_report_message(msg, composed_message)


    #// Function: issue
    #// Immediately issues the message which is currently being processed. This
    #// is useful if the message is being ~CAUGHT~ but should still be emitted.
    #//
    #// Issuing a message will update the report_server stats, possibly multiple
    #// times if the message is not ~CAUGHT~.
    def issue(self):
        composed_message = ""
        modded_msg = UVMReportCatcher.m_modified_report_message
        rs = modded_msg.get_report_server()

        if modded_msg.get_action() != UVM_NO_ACTION:
            # no need to compose when neither UVM_DISPLAY nor UVM_LOG is set
            if (modded_msg.get_action() & (UVM_LOG | UVM_DISPLAY)):
                composed_message = rs.compose_report_message(modded_msg)
            rs.execute_report_message(modded_msg, composed_message)


    in_catcher = 0
    #//process_all_report_catchers
    #//method called by report_server.report to process catchers
    #//
    @classmethod
    def process_all_report_catchers(cls, rm):
        _iter = UVMCallback("_iter")
        catcher = None  # uvm_report_catcher
        thrown = 1
        orig_severity = 0  # uvm_severity
        #cls.in_catcher = 0  # static bit in_catcher
        l_report_object = rm.get_report_object()  # uvm_report_object

        if cls.in_catcher == 1:
            return 1

        cls.in_catcher = 1
        UVMCallbacksBase.m_tracing = 0  # turn off cb tracing so catcher stuff doesn't print

        orig_severity = rm.get_severity()  # cast to 'uvm_severity' removed
        cls.m_modified_report_message = rm

        catcher = UVMReportCb.get_first(_iter, l_report_object)
        if catcher is not None:
            if cls.m_debug_flags & cls.DO_NOT_MODIFY:
                #process p = process::self(); // Keep random stability
                p = None
                randstate = ""
                if p is not None:
                    randstate = p.get_randstate()
                cls.m_orig_report_message = rm.clone()
                #sv.cast(m_orig_report_message, rm.clone()) # have to clone, rm can be extended type
                if p is not None:
                    p.set_randstate(randstate)

        while catcher is not None:
            prev_sev = 0  # uvm_severity

            if catcher.callback_mode() is False:
                catcher = UVMReportCb.get_next(_iter, l_report_object)
                continue


            prev_sev = cls.m_modified_report_message.get_severity()
            cls.m_set_action_called = False
            thrown = catcher.process_report_catcher()

            # Set the action to the default action for the new severity
            # if it is still at the default for the previous severity,
            # unless it was explicitly set.
            if (not cls.m_set_action_called and
                cls.m_modified_report_message.get_severity() != prev_sev and
                cls.m_modified_report_message.get_action() ==
                l_report_object.get_report_action(prev_sev, "*@&*^*^*#")):

                cls.m_modified_report_message.set_action(
                    l_report_object.get_report_action(cls.m_modified_report_message.get_severity(), "*@&*^*^*#"))

            if thrown == 0:
                if orig_severity == UVM_FATAL:
                    cls.m_caught_fatal += 1
                elif orig_severity == UVM_ERROR:
                    cls.m_caught_error += 1
                elif orig_severity == UVM_WARNING:
                    cls.m_caught_warning += 1
                break
            catcher = UVMReportCb.get_next(_iter,l_report_object)

        # update counters if message was returned with demoted severity
        if orig_severity == UVM_FATAL:
            if cls.m_modified_report_message.get_severity() < orig_severity:
                cls.m_demoted_fatal += 1
        elif orig_severity == UVM_ERROR:
            if cls.m_modified_report_message.get_severity() < orig_severity:
                cls.m_demoted_error += 1
        elif orig_severity == UVM_WARNING:
            if cls.m_modified_report_message.get_severity() < orig_severity:
                cls.m_demoted_warning += 1

        cls.in_catcher = 0
        UVMCallbacksBase.m_tracing = 1  # turn tracing stuff back on
        return thrown


    #//process_report_catcher
    #//internal method to call user <catch()> method
    #//
    def process_report_catcher(self):
        act = self.catch()

        if act == UNKNOWN_ACTION:
            self.uvm_report_error("RPTCTHR", "uvm_report_self.catch() in catcher instance "
                + self.get_name() + " must return THROW or CAUGHT", UVM_NONE)

        cls = UVMReportCatcher
        if cls.m_debug_flags & cls.DO_NOT_MODIFY:
            cls.m_modified_report_message.copy(cls.m_orig_report_message)

        if act == CAUGHT and not (cls.m_debug_flags & cls.DO_NOT_CATCH):
            return 0
        return 1

    #// Function: summarize
    #//
    #// This function is called automatically by `UVMReportServer.report_summarize()`.
    #// It prints the statistics for the active catchers.
    @classmethod
    def summarize(cls):
        # s = ""
        q = []
        if UVMReportCatcher.do_report:
            q.append("\n--- UVM Report catcher Summary ---\n\n\n")
            q.append("Number of demoted UVM_FATAL reports  :{}\n".format(cls.m_demoted_fatal))
            q.append("Number of demoted UVM_ERROR reports  :{}\n".format(cls.m_demoted_error))
            q.append("Number of demoted UVM_WARNING reports:{}\n".format(cls.m_demoted_warning))
            q.append("Number of caught UVM_FATAL reports   :{}\n".format(cls.m_caught_fatal))
            q.append("Number of caught UVM_ERROR reports   :{}\n".format(cls.m_caught_error))
            q.append("Number of caught UVM_WARNING reports :{}\n".format(cls.m_caught_warning))
            from .uvm_root import uvm_top
            uvm_info_context("UVM/REPORT/CATCHER", UVM_STRING_QUEUE_STREAMING_PACK(q),UVM_LOW,uvm_top)



#//
#//--------------------------------------------------------------
#//    Copyright 2004-2009 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2019-2020 Tuomas Poikela
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#//--------------------------------------------------------------
"""
Title: Generic Register Operation Descriptors

This section defines the abstract register transaction item. It also defines
a descriptor for a physical bus operation that is used by <uvm_reg_adapter>
subtypes to convert from a protocol-specific address/data/rw operation to
a bus-independent, canonical r/w operation.
"""


from ..seq.uvm_sequence_item import UVMSequenceItem
from ..macros.uvm_object_defines import uvm_object_utils
from ..macros.uvm_message_defines import uvm_error, uvm_fatal
from .uvm_reg_model import (UVM_READ, UVM_NOT_OK, UVM_MEM, UVM_FRONTDOOR,
    UVM_ACCESS_NAMES, UVM_ELEMENT_KIND_NAMES)
from ..base.uvm_object_globals import UVM_HIGH, UVM_INFO
from ..base.sv import sv
from ..base.uvm_globals import uvm_report_enabled
from typing import TypeVar, List

_T0 = TypeVar('_T0')


class UVMRegItem(UVMSequenceItem):
    """
    CLASS: uvm_reg_item

    Defines an abstract register transaction item. No bus-specific information
    is present, although a handle to a <uvm_reg_map> is provided in case a user
    wishes to implement a custom address translation algorithm.
    """

    #  // Variable: element_kind
    #  //
    #  // Kind of element being accessed: REG, MEM, or FIELD. See <uvm_elem_kind_e>.
    #  //
    #  uvm_elem_kind_e element_kind


    #  // TODO: parameterize
    #  constraint max_values { value.size() > 0 && value.size() < 1000; }
    #
    #  // Variable: offset
    #  //
    #  // For memory accesses, the offset address. For bursts,
    #  // the ~starting~ offset address.
    #  //
    #  rand uvm_reg_addr_t offset

    #  // Variable: status
    #  //
    #  // The result of the transaction: IS_OK, HAS_X, or ERROR.
    #  // See <uvm_status_e>.
    #  //
    #  uvm_status_e status

    #  // Variable: local_map
    #  //
    #  // The local map used to obtain addresses. Users may customize
    #  // address-translation using this map. Access to the sequencer
    #  // and bus adapter can be obtained by getting this map's root map,
    #  // then calling <uvm_reg_map::get_sequencer> and
    #  // <uvm_reg_map::get_adapter>.
    #  //
    #  uvm_reg_map local_map



    #  // Variable: parent
    #  //
    #  // The sequence from which the operation originated.
    #  //
    #  rand uvm_sequence_base parent

    #  // Variable: prior
    #  //
    #  // The priority requested of this transfer, as defined by
    #  // <uvm_sequence_base::start_item>.
    #  //
    #  int prior = -1

    #  // Variable: extension
    #  //
    #  // Handle to optional user data, as conveyed in the call to
    #  // write(), read(), mirror(), or update() used to trigger the operation.
    #  //
    #  rand uvm_object extension

    #  // Variable: bd_kind
    #  //
    #  // If path is UVM_BACKDOOR, this member specifies the abstraction
    #  // kind for the backdoor access, e.g. "RTL" or "GATES".
    #  //
    #  string bd_kind

    #  // Variable: fname
    #  //
    #  // The file name from where this transaction originated, if provided
    #  // at the call site.
    #  //
    #  string fname

    #  // Variable: lineno
    #  //
    #  // The file name from where this transaction originated, if provided
    #  // at the call site.
    #  //
    #  int lineno

    def __init__(self, name="") -> None:
        """
          Function: new

          Create a new instance of this type, giving it the optional `name`.

        Args:
            name:
        """
        UVMSequenceItem.__init__(self, name)
        #  // Variable: value
        #  //
        #  // The value to write to, or after completion, the value read from the DUT.
        #  // Burst operations use the <values> property.
        #  //
        self.value: List[int] = [0]  # rand uvm_reg_data_t value[]

        #  // Variable: path
        #  //
        #  // The path being used: <UVM_FRONTDOOR> or <UVM_BACKDOOR>.
        #  //
        #  uvm_path_e path
        self.path = UVM_FRONTDOOR

        self.status = 0
        self.fname = ""
        self.lineno = 0
        self.bd_kind = ""
        self.prior = -1
        self.extension = None
        self.parent = None
        self.offset = 0
        #  // Variable: kind
        #  //
        #  // Kind of access: READ or WRITE.
        #  //
        #  rand uvm_access_e kind
        self.kind = UVM_READ

        #  // Variable: element
        #  //
        #  // A handle to the RegModel model element associated with this transaction.
        #  // Use <element_kind> to determine the type to cast  to: <uvm_reg>,
        #  // <uvm_mem>, or <uvm_reg_field>.
        #  //
        self.element = None
        self.element_kind = -1

        #  // Variable: map
        #  //
        #  // The original map specified for the operation. The actual <map>
        #  // used may differ when a test or sequence written at the block
        #  // level is reused at the system level.
        #  //
        self.map = None

        self.local_map = None

    def convert2string(self) -> str:
        """
        Function: convert2string

        Returns a string showing the contents of this transaction.

        Returns:
            str: Reg item as string.
        """
        value_s = ""
        ele_name = "null"

        if self.element is not None:
            ele_name = self.element.get_full_name()

        kind_str         = UVM_ACCESS_NAMES[self.kind]
        element_kind_str = UVM_ELEMENT_KIND_NAMES[self.element_kind]

        s = ("kind="      + kind_str +
             " ele_kind=" + element_kind_str +
             " ele_name=" + ele_name)

        if (len(self.value) > 1 and uvm_report_enabled(UVM_HIGH, UVM_INFO, "RegModel")):
            value_s = "'{"
            for i in range(len(self.value)):
                value_s = value_s + sv.sformatf("%0h,", self.value[i])
            value_s = value_s[:-1] + "}"
        else:
            value_s = sv.sformatf("%0h", self.value[0])
        s = s + " value=" + value_s
        
        if self.element_kind == UVM_MEM:
            s = s + sv.sformatf(" offset=%0h", self.offset)

        map_name = "null"
        if self.map is not None:
            map_name = self.map.get_full_name()
        s = s + " map=" + map_name + " path=" + str(self.path)
        s = s + " status=" + str(self.status)
        return s


    def do_copy(self, rhs) -> None:
        """
        Function: do_copy

        Copy the ~rhs~ object into this object. The ~rhs~ object must
        derive from `UVMRegItem`.
        """
        if rhs is None:
            uvm_fatal("REG/NULL","do_copy: rhs argument is null")

        arr_rhs_ = []
        if not sv.cast(arr_rhs_, rhs, UVMRegItem):
            uvm_error("WRONG_TYPE","Provided rhs is not of type uvm_reg_item")
            return
        rhs_ = arr_rhs_[0]

        super().copy(rhs)
        self.element_kind = rhs_.element_kind
        self.element = rhs_.element
        self.kind = rhs_.kind
        self.value = rhs_.value
        self.offset = rhs_.offset
        self.status = rhs_.status
        self.local_map = rhs_.local_map
        self.map = rhs_.map
        self.path = rhs_.path
        self.extension = rhs_.extension
        self.bd_kind = rhs_.bd_kind
        self.parent = rhs_.parent
        self.prior = rhs_.prior
        self.fname = rhs_.fname
        self.lineno = rhs_.lineno



uvm_object_utils(UVMRegItem)


class UVMRegBusOp():
    """
    CLASS: UVMRegBusOp

    Class that defines a generic bus transaction for register and memory accesses, having
    ~kind~ (read or write), ~address~, ~data~, and ~byte enable~ information.
    If the bus is narrower than the register or memory location being accessed,
    there will be multiple of these bus operations for every abstract
    `UVMRegItem` transaction. In this case, ~data~ represents the portion
    of <uvm_reg_item::value> being transferred during this bus cycle.
    If the bus is wide enough to perform the register or memory operation in
    a single cycle, ~data~ will be the same as <uvm_reg_item::value>.
    """

    #typedef struct {
    #
    #  // Variable: kind
    #  //
    #  // Kind of access: READ or WRITE.
    #  //
    #  uvm_access_e kind
    #
    #
    #  // Variable: addr
    #  //
    #  // The bus address.
    #  //
    #  uvm_reg_addr_t addr
    #
    #
    #  // Variable: data
    #  //
    #  // The data to write. If the bus width is smaller than the register or
    #  // memory width, ~data~ represents only the portion of ~value~ that is
    #  // being transferred this bus cycle.
    #  //
    #  uvm_reg_data_t data
    #
    #
    #  // Variable: n_bits
    #  //
    #  // The number of bits of <uvm_reg_item::value> being transferred by
    #  // this transaction.
    #
    #  int n_bits
    #
    #  /*
    #  constraint valid_n_bits {
    #     n_bits > 0
    #     n_bits <= `UVM_REG_DATA_WIDTH
    #  }
    #  */
    #
    #
    #  // Variable: byte_en
    #  //
    #  // Enables for the byte lanes on the bus. Meaningful only when the
    #  // bus supports byte enables and the operation originates from a field
    #  // write/read.
    #  //
    #  uvm_reg_byte_en_t byte_en
    #
    #
    #  // Variable: status
    #  //
    #  // The result of the transaction: UVM_IS_OK, UVM_HAS_X, UVM_NOT_OK.
    #  // See <uvm_status_e>.
    #  //
    #  uvm_status_e status
    #
    #} uvm_reg_bus_op

    def __init__(self) -> None:
        """
        Constructor
        """
        self.kind = UVM_READ
        self.addr = 0
        self.data = 0
        self.n_bits = 0
        self.byte_en = 0x0
        self.status = UVM_NOT_OK



#// 
#// -------------------------------------------------------------
#//    Copyright 2004-2008 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    All Rights Reserved Worldwide
#// 
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#// 
#//        http://www.apache.org/licenses/LICENSE-2.0
#// 
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#// 
#
#//------------------------------------------------------------------------------
#// Title: Shared Register and Memory Access Test Sequences
#//------------------------------------------------------------------------------
#// This section defines sequences for testing registers and memories that are
#// shared between two or more physical interfaces, i.e. are associated with
#// more than one <uvm_reg_map> instance.
#//------------------------------------------------------------------------------
#
#
#//------------------------------------------------------------------------------
#// Class: uvm_reg_shared_access_seq
#//
#// Verify the accessibility of a shared register
#// by writing through each address map
#// then reading it via every other address maps
#// in which the register is readable and the backdoor,
#// making sure that the resulting value matches the mirrored value.
#//
#// If bit-type resource named
#// "NO_REG_TESTS" or "NO_REG_SHARED_ACCESS_TEST"
#// in the "REG::" namespace
#// matches the full name of the register,
#// the register is not tested.
#//
#//| uvm_resource_db#(bit)::set({"REG::",regmodel.blk.r0.get_full_name()},
#//|                            "NO_REG_TESTS", 1, this);
#//
#// Registers that contain fields with unknown access policies
#// cannot be tested.
#//
#// The DUT should be idle and not modify any register during this test.
#//
#//------------------------------------------------------------------------------
#
#class uvm_reg_shared_access_seq extends uvm_reg_sequence #(uvm_sequence #(uvm_reg_item));
#
#   // Variable: rg
#   // The register to be tested
#   uvm_reg rg;
#
#   `uvm_object_utils(uvm_reg_shared_access_seq)
#
#   function new(string name="uvm_reg_shared_access_seq");
#     super.new(name);
#   endfunction
#
#
#   virtual task body();
#      uvm_reg_data_t  other_mask;
#      uvm_reg_data_t  wo_mask[$];
#      uvm_reg_field fields[$];
#      uvm_reg_map maps[$];
#
#      if (rg == null) begin
#         `uvm_error("uvm_reg_shared_access_seq", "No register specified to run sequence on");
#         return;
#      end
#
#      // Registers with some attributes are not to be tested
#      if (uvm_resource_db#(bit)::get_by_name({"REG::",rg.get_full_name()},
#                                             "NO_REG_TESTS", 0) != null ||
#          uvm_resource_db#(bit)::get_by_name({"REG::",rg.get_full_name()},
#                                             "NO_REG_SHARED_ACCESS_TEST", 0) != null )
#        return;
#
#      // Only look at shared registers
#      if (rg.get_n_maps() < 2) return;
#      rg.get_maps(maps);
#
#      // Let's see what kind of bits we have...
#      rg.get_fields(fields);
#
#      // Identify unpredictable bits and the ones we shouldn't change
#      other_mask = 0;
#      foreach (fields[k]) begin
#         int lsb, w;
#         
#         lsb = fields[k].get_lsb_pos();
#         w   = fields[k].get_n_bits();
#         
#         if (!fields[k].is_known_access(maps[0])) begin
#            repeat (w) begin
#               other_mask[lsb++] = 1'b1;
#            end
#         end
#      end
#      
#      // WO bits will always readback as 0's but the mirror
#      // with return what is supposed to have been written
#      // so we cannot use the mirror-check function
#      foreach (maps[j]) begin
#         uvm_reg_data_t  wo;
#         wo = 0;
#         foreach (fields[k]) begin
#            int lsb, w;
#            
#            lsb = fields[k].get_lsb_pos();
#            w   = fields[k].get_n_bits();
#            
#            if (fields[k].get_access(maps[j]) == "WO") begin
#               repeat (w) begin
#                  wo[lsb++] = 1'b1;
#               end
#            end
#         end
#         wo_mask[j] = wo;
#      end
#      
#      // Try to write through each map
#      foreach (maps[j]) begin
#         uvm_status_e status;
#         uvm_reg_data_t  prev, v;
#         
#         // The mirror should contain the initial value
#         prev = rg.get();
#         
#         // Write a random value, except in those "don't touch" fields
#         v = ({$random, $random} & ~other_mask) | (prev & other_mask);
#         
#         `uvm_info("uvm_reg_shared_access_seq", $sformatf("Writing register %s via map \"%s\"...",
#                                    rg.get_full_name(), maps[j].get_full_name), UVM_LOW);
#         
#         `uvm_info("uvm_reg_shared_access_seq", $sformatf("Writing 'h%h over 'h%h", v, prev),UVM_DEBUG);
#         
#         rg.write(status, v, UVM_FRONTDOOR, maps[j], this);
#         if (status != UVM_IS_OK) begin
#            `uvm_error("uvm_reg_shared_access_seq", $sformatf("Status was %s when writing register \"%s\" through map \"%s\".",
#                                        status.name(), rg.get_full_name(), maps[j].get_full_name()));
#         end
#         
#         foreach (maps[k]) begin
#            uvm_reg_data_t  actual, exp;
#            
#            `uvm_info("uvm_reg_shared_access_seq", $sformatf("Reading register %s via map \"%s\"...",
#                                       rg.get_full_name(), maps[k].get_full_name()), UVM_LOW);
#            
#            // Was it what we expected?
#            exp = rg.get() & ~wo_mask[k];
#            
#            rg.read(status, actual, UVM_FRONTDOOR, maps[k], this);
#            if (status != UVM_IS_OK) begin
#               `uvm_error("uvm_reg_shared_access_seq", $sformatf("Status was %s when reading register \"%s\" through map \"%s\".",
#                                           status.name(), rg.get_full_name(), maps[k].get_full_name()));
#            end
#            
#            `uvm_info("uvm_reg_shared_access_seq", $sformatf("Read 'h%h, expecting 'h%h",
#                                        actual, exp),UVM_DEBUG);
#            
#            if (actual !== exp) begin
#               `uvm_error("uvm_reg_shared_access_seq", $sformatf("Register \"%s\" through map \"%s\" is 'h%h instead of 'h%h after writing 'h%h via map \"%s\" over 'h%h.",
#                                           rg.get_full_name(), maps[k].get_full_name(),
#                                           actual, exp, v, maps[j].get_full_name(), prev));
#            end
#         end
#      end
#   endtask: body
#endclass: uvm_reg_shared_access_seq
#
#
#//------------------------------------------------------------------------------
#// Class: uvm_mem_shared_access_seq
#//------------------------------------------------------------------------------
#//
#// Verify the accessibility of a shared memory
#// by writing through each address map
#// then reading it via every other address maps
#// in which the memory is readable and the backdoor,
#// making sure that the resulting value matches the written value.
#//
#// If bit-type resource named
#// "NO_REG_TESTS", "NO_MEM_TESTS",
#// "NO_REG_SHARED_ACCESS_TEST" or "NO_MEM_SHARED_ACCESS_TEST"
#// in the "REG::" namespace
#// matches the full name of the memory,
#// the memory is not tested.
#//
#//| uvm_resource_db#(bit)::set({"REG::",regmodel.blk.mem0.get_full_name()},
#//|                            "NO_MEM_TESTS", 1, this);
#//
#// The DUT should be idle and not modify the memory during this test.
#//
#//------------------------------------------------------------------------------
#
#class uvm_mem_shared_access_seq extends uvm_reg_sequence #(uvm_sequence #(uvm_reg_item));
#
#   // variable: mem
#   // The memory to be tested
#   uvm_mem mem;
#
#   `uvm_object_utils(uvm_mem_shared_access_seq)
#
#   function new(string name="uvm_mem_shared_access_seq");
#     super.new(name);
#   endfunction
#
#   virtual task body();
#      int read_from;
#      uvm_reg_map maps[$];
#
#      if (mem == null) begin
#         `uvm_error("uvm_mem_shared_access_seq", "No memory specified to run sequence on");
#         return;
#      end
#
#      // Memories with some attributes are not to be tested
#      if (uvm_resource_db#(bit)::get_by_name({"REG::",mem.get_full_name()},
#                                             "NO_REG_TESTS", 0) != null ||
#          uvm_resource_db#(bit)::get_by_name({"REG::",mem.get_full_name()},
#                                             "NO_MEM_TESTS", 0) != null ||
#          uvm_resource_db#(bit)::get_by_name({"REG::",mem.get_full_name()},
#                                             "NO_REG_SHARED_ACCESS_TEST", 0) != null ||
#          uvm_resource_db#(bit)::get_by_name({"REG::",mem.get_full_name()},
#                                             "NO_MEM_SHARED_ACCESS_TEST", 0) != null )
#            return;
#
#      // Only look at shared memories
#      if (mem.get_n_maps() < 2) return;
#      mem.get_maps(maps);
#
#      // We need at least a backdoor or a map that can read
#      // the shared memory
#      read_from = -1;
#      if (mem.get_backdoor() == null) begin
#         foreach (maps[j]) begin
#            string right;
#            right = mem.get_access(maps[j]);
#            if (right == "RW" ||
#                right == "RO") begin
#               read_from = j;
#               break;
#            end
#         end
#         if (read_from < 0) begin
#            `uvm_warning("uvm_mem_shared_access_seq", $sformatf("Memory \"%s\" cannot be read from any maps or backdoor. Shared access not verified.", mem.get_full_name()));
#            return;
#         end
#      end
#      
#      // Try to write through each map
#      foreach (maps[j]) begin
#         
#         `uvm_info("uvm_mem_shared_access_seq", $sformatf("Writing shared memory \"%s\" via map \"%s\".",
#                                    mem.get_full_name(), maps[j].get_full_name()), UVM_LOW);
#         
#         // All addresses
#         for (int offset = 0; offset < mem.get_size(); offset++) begin
#            uvm_status_e status;
#            uvm_reg_data_t  prev, v;
#            
#            // Read the initial value
#            if (mem.get_backdoor() != null) begin
#               mem.peek(status, offset, prev);
#               if (status != UVM_IS_OK) begin
#                  `uvm_error("uvm_mem_shared_access_seq", $sformatf("Status was %s when reading initial value of \"%s\"[%0d] through backdoor.",
#                                              status.name(), mem.get_full_name(), offset));
#               end
#            end
#            else begin
#               mem.read(status, offset, prev, UVM_FRONTDOOR, maps[read_from], this);
#               if (status != UVM_IS_OK) begin
#                  `uvm_error("uvm_mem_shared_access_seq", $sformatf("Status was %s when reading initial value of \"%s\"[%0d] through map \"%s\".",
#                                              status.name(), mem.get_full_name(),
#                                              offset, maps[read_from].get_full_name()));
#               end
#            end
#            
#            
#            // Write a random value,
#            v = {$random, $random};
#            
#            mem.write(status, offset, v, UVM_FRONTDOOR, maps[j], this);
#            if (status != UVM_IS_OK) begin
#               `uvm_error("uvm_mem_shared_access_seq", $sformatf("Status was %s when writing \"%s\"[%0d] through map \"%s\".",
#                                           status.name(), mem.get_full_name(), offset, maps[j].get_full_name()));
#            end
#            
#            // Read back from all other maps
#            foreach (maps[k]) begin
#               uvm_reg_data_t  actual, exp;
#               
#               mem.read(status, offset, actual, UVM_FRONTDOOR, maps[k], this);
#               if (status != UVM_IS_OK) begin
#                  `uvm_error("uvm_mem_shared_access_seq", $sformatf("Status was %s when reading %s[%0d] through map \"%s\".",
#                                              status.name(), mem.get_full_name(), offset, maps[k].get_full_name()));
#               end
#               
#               // Was it what we expected?
#               exp = v;
#               if (mem.get_access(maps[j]) == "RO") begin
#                  exp = prev;
#               end
#               if (mem.get_access(maps[k]) == "WO") begin
#                  exp = 0;
#               end
#               // Trim to number of bits
#               exp &= (1 << mem.get_n_bits()) - 1;
#               if (actual !== exp) begin
#                  `uvm_error("uvm_mem_shared_access_seq", $sformatf("%s[%0d] through map \"%s\" is 'h%h instead of 'h%h after writing 'h%h via map \"%s\" over 'h%h.",
#                                              mem.get_full_name(), offset, maps[k].get_full_name(),
#                                              actual, exp, v, maps[j].get_full_name(), prev));
#               end
#            end
#         end
#      end
#   endtask: body
#endclass: uvm_mem_shared_access_seq
#
#
#
#//------------------------------------------------------------------------------
#// Class: uvm_reg_mem_shared_access_seq
#//------------------------------------------------------------------------------
#//
#// Verify the accessibility of all shared registers
#// and memories in a block
#// by executing the <uvm_reg_shared_access_seq>
#// and <uvm_mem_shared_access_seq>
#// sequence respectively on every register and memory within it.
#//
#// If bit-type resource named
#// "NO_REG_TESTS", "NO_MEM_TESTS",
#// "NO_REG_SHARED_ACCESS_TEST" or "NO_MEM_SHARED_ACCESS_TEST"
#// in the "REG::" namespace
#// matches the full name of the block,
#// the block is not tested.
#//
#//| uvm_resource_db#(bit)::set({"REG::",regmodel.blk.get_full_name(),".*"},
#//|                            "NO_REG_TESTS", 1, this);
#//
#//------------------------------------------------------------------------------
#
#class uvm_reg_mem_shared_access_seq extends uvm_reg_sequence #(uvm_sequence #(uvm_reg_item));
#
#   // Variable: model
#   //
#   // The block to be tested
#   //
#   //| uvm_reg_block model; 
#
#
#   // Variable: reg_seq
#   //
#   // The sequence used to test one register
#   //
#   protected uvm_reg_shared_access_seq reg_seq;
#   
#
#   // Variable: mem_seq
#   //
#   // The sequence used to test one memory
#   //
#   protected uvm_mem_shared_access_seq mem_seq;
#   
#   `uvm_object_utils(uvm_reg_mem_shared_access_seq)
#
#   function new(string name="uvm_reg_mem_shared_access_seq");
#     super.new(name);
#   endfunction
#
#
#   // Task: body
#   //
#   // Executes the Shared Register and Memory sequence
#   //
#   virtual task body();
#
#      if (model == null) begin
#         `uvm_error("uvm_reg_mem_shared_access_seq", "No register model specified to run sequence on");
#         return;
#      end
#      
#      uvm_report_info("STARTING_SEQ",{"\n\nStarting ",get_name()," sequence...\n"},UVM_LOW);
#
#      reg_seq = uvm_reg_shared_access_seq::type_id::create("reg_shared_access_seq");
#      mem_seq = uvm_mem_shared_access_seq::type_id::create("reg_shared_access_seq");
#
#      this.reset_blk(model);
#      model.reset();
#
#      do_block(model);
#   endtask: body
#
#
#   // Task: do_block
#   //
#   // Test all of the registers and memories in a block
#   //
#   protected virtual task do_block(uvm_reg_block blk);
#      uvm_reg regs[$];
#      uvm_mem mems[$];
#      
#      if (uvm_resource_db#(bit)::get_by_name({"REG::",blk.get_full_name()},
#                                             "NO_REG_TESTS", 0) != null ||
#          uvm_resource_db#(bit)::get_by_name({"REG::",blk.get_full_name()},
#                                             "NO_MEM_TESTS", 0) != null ||
#          uvm_resource_db#(bit)::get_by_name({"REG::",blk.get_full_name()},
#                                             "NO_REG_SHARED_ACCESS_TEST", 0) != null ||
#          uvm_resource_db#(bit)::get_by_name({"REG::",blk.get_full_name()},
#                                             "NO_MEM_SHARED_ACCESS_TEST", 0) != null )
#        return;
#
#      this.reset_blk(model);
#      model.reset();
#
#      // Iterate over all registers, checking accesses
#      blk.get_registers(regs, UVM_NO_HIER);
#      foreach (regs[i]) begin
#         // Registers with some attributes are not to be tested
#         if (uvm_resource_db#(bit)::get_by_name({"REG::",regs[i].get_full_name()},
#                                                "NO_REG_TESTS", 0) != null ||
#             uvm_resource_db#(bit)::get_by_name({"REG::",regs[i].get_full_name()},
#                                                "NO_REG_SHARED_ACCESS_TEST", 0) != null )
#           continue;
#         reg_seq.rg = regs[i];
#         reg_seq.start(this.get_sequencer(), this);
#      end
#
#      // Iterate over all memories, checking accesses
#      blk.get_memories(mems, UVM_NO_HIER);
#      foreach (mems[i]) begin
#         // Registers with some attributes are not to be tested
#         if (uvm_resource_db#(bit)::get_by_name({"REG::",mems[i].get_full_name()},
#                                                "NO_REG_TESTS", 0) != null ||
#             uvm_resource_db#(bit)::get_by_name({"REG::",mems[i].get_full_name()},
#                                                "NO_MEM_TESTS", 0) != null ||
#             uvm_resource_db#(bit)::get_by_name({"REG::",mems[i].get_full_name()},
#                                                "NO_REG_SHARED_ACCESS_TEST", 0) != null ||
#             uvm_resource_db#(bit)::get_by_name({"REG::",mems[i].get_full_name()},
#                                                "NO_MEM_SHARED_ACCESS_TEST", 0) != null )
#            continue;
#         mem_seq.mem = mems[i];
#         mem_seq.start(this.get_sequencer(), this);
#      end
#
#      begin
#         uvm_reg_block blks[$];
#         
#         blk.get_blocks(blks);
#         foreach (blks[i]) begin
#            do_block(blks[i]);
#         end
#      end
#   endtask: do_block
#
#
#   //
#   // task: reset_blk
#   // Reset the DUT that corresponds to the specified block abstraction class.
#   //
#   // Currently empty.
#   // Will rollback the environment's phase to the ~reset~
#   // phase once the new phasing is available.
#   //
#   // In the meantime, the DUT should be reset before executing this
#   // test sequence or this method should be implemented
#   // in an extension to reset the DUT.
#   //
#   virtual task reset_blk(uvm_reg_block blk);
#   endtask
#
#
#endclass: uvm_reg_mem_shared_access_seq
#



#
#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

# About: hierarchy
# This test is supposed to test the hierarchical connection
# between ports of hierarchical threads using a uvm_tlm_fifo exports.
#
# Walk through the test:
# ======================
# 1. A thread *gen* will use a uvm_blocking_put_port to put a transaction.
#
# 2. A *conv* thread will use a uvm_blocking_put_port to put a transaction,an
# uvm_blocking_get_port to get a transactions, and a uvm_analysis_ port to a
# write a transaction through it.
#
# 3. Another thread *bfm* will use a uvm_blocking_get_port to only get the
# transactions that has been sent by the other threads.
#
# 4. A *listener* will extends the uvm_subscriber, to implement the analysis port
# write function.
#
# 5. A *producer* component will use uvm_blocking_put_port and uvm_analysis_port and
# a uvm_tlm_fifo, to connect the *gen* and *conv* put, get and analysis ports
# through the fifo exports.
#
# 6. A *consumer* component will use uvm_blocking_put export and connect it directly
# to a fifo export, also will connect the *bfm*  get port to the fifo export
#
# 7. At *top* env, the *producer*, *consumer*, and the *listener* will be connected
#

import cocotb
from cocotb.triggers import Timer

from uvm import *

#  //----------------------------------------------------------------------
#  // class transaction
#  //----------------------------------------------------------------------


class transaction(UVMTransaction):


    def __init__(self, name='trans'):
        super().__init__(name)
        self.addr = 0
        self.data = 0
        self.rand("data", range(1 << 32))
        self.rand("addr", range(1 << 32))

    def do_copy(self, t):
        self.data = t.data
        self.addr = t.addr
        #    endfunction

    def do_compare(self, a, b):
        return a.data == b.data and a.addr == b.addr

    def do_clone(self):
        t = transaction()
        t.copy(self)
        return t

    def convert2string(self):
        s = sv.sformatf("[ addr = %s, data = %s ]", hex(self.addr), hex(self.data))
        return s
    #  endclass
uvm_object_utils(transaction)

#  //----------------------------------------------------------------------
#  // component gen
#  //----------------------------------------------------------------------

class gen(UVMComponent):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.put_port = UVMBlockingPutPort("put_port", self)
        self.num_items = 0

    
    async def run_phase(self, phase):
        for i in range(20):
            t = transaction()
            if not t.randomize():
                uvm_error("GEN", "Failed to randomize transaction")
            uvm_info("gen", sv.sformatf("sending  : %s", t.convert2string()), UVM_MEDIUM)
            await self.put_port.put(t)
            self.num_items += 1


#  //----------------------------------------------------------------------
#  // componenet conv
#  //----------------------------------------------------------------------

class conv(UVMComponent):


    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.put_port = UVMBlockingPutPort("put_port", self)
        self.get_port = UVMBlockingGetPort("get_port", self)
        self.ap = UVMAnalysisPort("analysis_port", self)

    
    async def run_phase(self, phase):
        while True:
            t = []
            await self.get_port.get(t)
            t = t[0]
            self.ap.write(t)
            await self.put_port.put(t)

    #  endclass

#
#  //----------------------------------------------------------------------
#  // componenet bfm
#  //----------------------------------------------------------------------

class bfm(UVMComponent):
    #
    #    uvm_blocking_get_port #(transaction) get_port
    #
    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.get_port = UVMBlockingGetPort("get_port", self)

    
    async def run_phase(self, phase):

        while True:
            t = []
            await self.get_port.get(t)
            t = t[0]
            uvm_info("bfm", sv.sformatf("receiving: %s", t.convert2string()),UVM_MEDIUM)

    #  endclass


class listener(UVMSubscriber): #(transaction)
    def __init__(self, name, parent):
        super().__init__(name,parent)
        self.num_items = 0

    def write(self, t):
        uvm_info(self.m_name, sv.sformatf("Received: %s", t.convert2string()), UVM_MEDIUM)
        self.num_items += 1


#  //----------------------------------------------------------------------
#  // componenet producer
#  //----------------------------------------------------------------------
#  // begin codeblock producer
class producer(UVMComponent):
    #
    #    uvm_blocking_put_port #(transaction) put_port
    #    uvm_analysis_port #(transaction) ap
    #
    #    gen g
    #    conv c
    #    uvm_tlm_fifo #(transaction) f
    #
    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.put_port = UVMBlockingPutPort("put_port", self)
        self.ap = UVMAnalysisPort("analysis_port", self)
        self.g = gen("gen", self)
        self.c = conv("conv", self)
        self.f = UVMTLMFIFO("fifo", self)


    def connect_phase(self, phase):
        self.g.put_port.connect(self.f.blocking_put_export)  # A
        self.c.get_port.connect(self.f.blocking_get_export)  # B
        self.c.put_port.connect(self.put_port) # C
        self.c.ap.connect(self.ap)

    #  endclass
    #

#  //----------------------------------------------------------------------
#  // componenet consumer
#  //----------------------------------------------------------------------
class consumer(UVMComponent):
    #
    #    uvm_blocking_put_export #(transaction) put_export
    #
    #    bfm b
    #    uvm_tlm_fifo #(transaction) f
    #
    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.put_export = UVMBlockingPutExport("put_export", self)
        self.f = UVMTLMFIFO("fifo", self)
        self.b = bfm("bfm", self)

    def connect_phase(self, phase):
       self.put_export.connect(self.f.blocking_put_export)
       self.b.get_port.connect(self.f.blocking_get_export)


#  //----------------------------------------------------------------------
#  // componenet top
#  //----------------------------------------------------------------------

class top(UVMEnv):
    #
    #    producer p
    #    consumer c
    #    listener l
    #
    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.p = producer("producer", self)
        self.c = consumer("consumer", self)
        self.list = listener("listener", self)

        # Connections may also be done in the constructor, if you wish
        self.p.put_port.connect(self.c.put_export)
        self.p.ap.connect(self.list.analysis_export)

    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        uvm_info("ENV_TOP", "run_phase started", UVM_MEDIUM)
        await Timer(100, "NS")
        phase.drop_objection(self)


    def check_phase(self, phase):
        if self.list.num_items == 0:
            uvm_fatal("TOP_ENV", "Listener got 0 items")


#  //----------------------------------------------------------------------
#  // environment env
#  //----------------------------------------------------------------------
class env(UVMEnv):

    def __init__(self, name="env"):
        super().__init__(name)
        self.t = top("top", self)


    async def run_phase(self, phase):
        phase.raise_objection(self)
        await Timer(900, "NS")
        phase.drop_objection(self)


#  //----------------------------------------------------------------------
#  // module test
#  //----------------------------------------------------------------------
@cocotb.test()
async def module_top(dut):
    e = env("e")
    await run_test()




import unittest

from uvm.base.uvm_scope_stack import UVMScopeStack


class TestUVMScopeStack(unittest.TestCase):

    def test_depth(self):
        stack = UVMScopeStack()
        self.assertEqual(stack.depth(), 0)

    def test_down_and_up(self):
        stack = UVMScopeStack()
        stack.down('xxx')
        stack.down('yyy')
        stack.down('zzz')
        self.assertEqual(stack.get(), 'xxx.yyy.zzz')
        self.assertEqual(stack.depth(), 3)
        stack.up()
        self.assertEqual(stack.get(), 'xxx.yyy')

    def test_down_and_up_element(self):
        stack = UVMScopeStack()
        stack.down('uvm_top')
        stack.down_element(4)
        self.assertEqual(stack.get(), 'uvm_top[4]')
        self.assertEqual(stack.depth(), 2)
        stack.up_element()
        self.assertEqual(stack.get(), 'uvm_top')


if __name__ == '__main__':
    unittest.main()



#//
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2021 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.clock import Clock
from cocotb.triggers import Timer

from uvm import *

from tb_env import tb_env
from vip import *

from apb.apb_if import apb_if

# Set to True to allow VIPs (drivers/monitors) raise/drop objections
# tpoikela: TODO This is not working properly at the moment
hier_objection = False


class test(UVMTest):

    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.env = tb_env("env")
        self.env.hier_objection = hier_objection


    def start_of_simulation_phase(self, phase):
        cs_ = UVMCoreService.get()
        top = cs_.get_root()

    def check_phase(self, phase):
        self.error = self.env.has_errors()
        if self.error:
            uvm_fatal("TEST FAILED", "check_phase of test threw fatal")


uvm_component_utils(test)


class tb_ctl_if():

    def __init__(self, clk, sclk, rst, intr):
        self.clk = clk
        self.sclk = sclk
        self.rst = rst
        self.intr = intr


async def do_reset_and_start_clocks(dut):
    await Timer(100, "NS")
    dut.rst <= 1
    await Timer(200, "NS")
    dut.clk <= 0
    dut.sclk <= 0
    await Timer(500, "NS")
    cocotb.fork(Clock(dut.clk, 50, "NS").start())
    cocotb.fork(Clock(dut.sclk, 250, "NS").start())


@cocotb.test()
async def test_codec(dut):

    # Create the interfaces and bus map for APB
    apb_bus_map = {"clk": "clk", "psel": "apb_psel", "penable": "apb_penable",
            "pwrite": "apb_pwrite", "prdata": "apb_prdata", "pwdata": "apb_pwdata",
            "paddr": "apb_paddr"}
    apb_vif = apb_if(dut, bus_map=apb_bus_map, name="")

    vip0 = vip_if(
        clk=dut.sclk,
        Tx=dut.rx,
        Rx=dut.tx
    )

    ctl = tb_ctl_if(
        clk=dut.clk,
        sclk=dut.sclk,
        rst=dut.rst,
        intr=dut.intr
    )


    UVMConfigDb.set(None, "env", "vif", ctl)
    UVMConfigDb.set(None, "env.apb", "vif", apb_vif)
    UVMConfigDb.set(None, "env.vip", "vif", vip0)

    cocotb.fork(do_reset_and_start_clocks(dut))

    await run_test()



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------


# About: producer_consumer
# This test is the basic simple to illustrates how to create a producer and
# consumer which are completely independent and connecting them using a uvm_tlm_fifo channel
#
# Walk through the test:
#
# Two processes *producer* and *consumer* will use uvm_blocking_put port and
# uvm_blocking_get port, to transfer an integer from the producer to the
# consumer through fifo exports at the connection function of an environment
# class
#
# Note that unlike in SystemVerilog, in Python we use arrays instead of ref
# parameters to obtain values from  blocking (coroutine) functions.


import cocotb
from cocotb.triggers import Timer

from uvm.tlm1 import *
from uvm.base.uvm_component import UVMComponent
from uvm.base.uvm_globals import run_test
from uvm.base.uvm_object_globals import UVM_DEBUG, UVM_MEDIUM
from uvm.comps.uvm_env import UVMEnv
from uvm.macros import *

from uvm.base.uvm_debug import UVMDebug

# UVMDebug.full_debug()

#  //----------------------------------------------------------------------
#  // class producer
#  //----------------------------------------------------------------------
class producer(UVMComponent):

    def __init__(self, name, p=None):
        super().__init__(name,p)
        self.put_port = UVMBlockingPutPort("put_port", self)
        self.num_items = 0

    
    async def run_phase(self, phase):
        #
        randval = 0

        for i in range(10):
            randval = sv.random() % 100
            #10
            await Timer(10, "NS")
            uvm_info("producer", sv.sformatf("sending %d", randval), UVM_MEDIUM)
            await self.put_port.put(randval)
            self.num_items += 1
        uvm_info("PRODUCER", "Finished run_phase", UVM_LOW)


class consumer(UVMComponent):

    def __init__(self, name, p=None):
        super().__init__(name,p)
        self.get_port = UVMBlockingGetPort("get_port", self)
        self.num_items = 0

    
    async def run_phase(self, phase):
        val = 0
        while True:
            uvm_info("CONSUMER", "Started to get items..", UVM_MEDIUM)
            arr = []
            await self.get_port.get(arr)
            val = arr[0]
            uvm_info("consumer", sv.sformatf("receiving %d", val), UVM_MEDIUM)
            self.num_items += 1
        uvm_info("CONSUMER", "Finished run_phase", UVM_LOW)


#  //----------------------------------------------------------------------
#  // class env
#  //----------------------------------------------------------------------

class env(UVMEnv):

    def __init__(self, name="env"):
        super().__init__(name)
        self.p = producer("producer", self)
        self.c = consumer("consumer", self)
        self.f = UVMTLMFIFO("fifo", self)
        sv.display("fifo put_export: %s", self.f.m_name)


    def connect_phase(self, phase):
        self.p.put_port.connect(self.f.put_export)
        self.c.get_port.connect(self.f.get_export)
        self.f.set_report_verbosity_level_hier(UVM_DEBUG)

    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        await Timer(1000, "NS")
        phase.drop_objection(self)


    def check_phase(self, phase):
        super().check_phase(phase)
        if self.p.num_items != self.c.num_items:
            uvm_fatal("TEST_ERR", "Mismatch in number of items: producer " +
                    str(self.p.num_items) + ' cons: ' + str(self.c.num_items))

    # endclass


@cocotb.test()
async def module_top(dut):
    # Main body of module top:
    e = env("prod_cons_env")
    await run_test()



#//
#// -------------------------------------------------------------
#//    Copyright 2010 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from uvm.base.uvm_object import UVMObject
from uvm.base.uvm_pool import UVMObjectStringPool
from uvm.base.uvm_queue import UVMQueue
from uvm.macros import (uvm_error)

#//
#// CLASS: uvm_reg_file
#// Register file abstraction base class
#//
#// A register file is a collection of register files and registers
#// used to create regular repeated structures.
#//
#// Register files are usually instantiated as arrays.
#//

class UVMRegFile(UVMObject):

    #   local uvm_reg_block     parent
    #   local uvm_reg_file   m_rf
    #   local string            default_hdl_path = "RTL"
    #   local uvm_object_string_pool #(uvm_queue #(string)) hdl_paths_pool

    #   //----------------------
    #   // Group: Initialization
    #   //----------------------

    def __init__(self, name=""):
        """         
          
           Function: new
          
           Create a new instance
          
           Creates an instance of a register file abstraction class
           with the specified name.
          
          extern def                  new        (self,string name=""):
        Args:
            name: 
        """
        super().__init__(name)
        self.hdl_paths_pool = UVMObjectStringPool("hdl_paths", UVMQueue)
        self.parent = None  # uvm_reg_block
        self.m_rf = None  # uvm_reg_file
        self.default_hdl_path = "RTL"  # type: str
        #    self.l = None  # type: hdl_paths_poo


    def configure(self, blk_parent, regfile_parent, hdl_path=""):
        """         
          
           Function: configure
           Configure a register file instance
          
           Specify the parent block and register file of the register file
           instance.
           If the register file is instantiated in a block,
           `regfile_parent` is specified as `None`.
           If the register file is instantiated in a register file,
           `blk_parent` must be the block parent of that register file and
           `regfile_parent` is specified as that register file.
          
           If the register file corresponds to a hierarchical RTL structure,
           its contribution to the HDL path is specified as the `hdl_path`.
           Otherwise, the register file does not correspond to a hierarchical RTL
           structure (e.g. it is physically flattened) and does not contribute
           to the hierarchical HDL path of any contained registers.
          
          extern function void     configure  (uvm_reg_block blk_parent,
                                               uvm_reg_file regfile_parent,
                                               string hdl_path = "")
        Args:
            blk_parent: 
            regfile_parent: 
            hdl_path: 
        """
        if blk_parent is None:
            uvm_error("UVM/RFILE/CFG/NOBLK", 
                ("UVMRegFile::configure() called without a parent block for instance '"
                    + self.get_name() + "' of register file type '", self.get_type_name()
                    + "'."))
            return

        self.parent = blk_parent
        self.m_rf = regfile_parent
        self.add_hdl_path(hdl_path)


    def get_full_name(self):
        """         
          ---------------------
           Group: Introspection
          ---------------------

          
           Function: get_name
           Get the simple name
          
           Return the simple object name of self register file.
          

          
           Function: get_full_name
           Get the hierarchical name
          
           Return the hierarchal name of self register file.
           The base of the hierarchical name is the root block.
          
          extern virtual def string        get_full_name(self):
        Returns:
        """
        blk = None  # uvm_reg_block 
        get_full_name = self.get_name()
     
        # Is there a parent register file?
        if (self.m_rf is not None):
           return self.m_rf.get_full_name() + "." + get_full_name
     
        # No: then prepend the full name of the parent block (if any)
        if self.parent is None:
            return get_full_name
        get_full_name = self.parent.get_full_name() + "." + get_full_name
        return get_full_name
    #endfunction: get_full_name

    #
    #   //
    #   // Function: get_parent
    #   // Get the parent block
    #   //
    #   extern virtual def uvm_reg_block get_parent (self):

    def get_block(self):
        """         
          extern virtual def uvm_reg_block get_block  (self):
        Returns:
        """
        return self.parent

    #   //
    #   // Function: get_regfile
    #   // Get the parent register file
    #   //
    #   // Returns ~None~ if self register file is instantiated in a block.
    #   //
    #   extern virtual def uvm_reg_file  get_regfile     (self):
    #
    #
    #   //----------------
    #   // Group: Backdoor
    #   //----------------
    #
    #   //
    #   // Function:  clear_hdl_path
    #   // Delete HDL paths
    #   //
    #   // Remove any previously specified HDL path to the register file instance
    #   // for the specified design abstraction.
    #   //
    #   extern def void clear_hdl_path    (self,string kind = "RTL"):
    #

    def add_hdl_path(self, path, kind="RTL"):
        """         
          
           Function:  add_hdl_path
           Add an HDL path
          
           Add the specified HDL path to the register file instance for the specified
           design abstraction. This method may be called more than once for the
           same design abstraction if the register file is physically duplicated
           in the design abstraction
          
          extern def void add_hdl_path      (self,string path, string kind = "RTL"):
        Args:
            path: 
            kind: 
        """
        paths = self.hdl_paths_pool.get(kind)
        paths.push_back(path)

    #
    #   //
    #   // Function:   has_hdl_path
    #   // Check if a HDL path is specified
    #   //
    #   // Returns TRUE if the register file instance has a HDL path defined for the
    #   // specified design abstraction. If no design abstraction is specified,
    #   // uses the default design abstraction specified for the nearest
    #   // enclosing register file or block
    #   //
    #   // If no design abstraction is specified, the default design abstraction
    #   // for self register file is used.
    #   //
    #   extern def bit  has_hdl_path      (self,string kind = ""):
    #
    #   //
    #   // Function:  get_hdl_path
    #   // Get the incremental HDL path(s)
    #   //
    #   // Returns the HDL path(s) defined for the specified design abstraction
    #   // in the register file instance. If no design abstraction is specified, uses
    #   // the default design abstraction specified for the nearest enclosing
    #   // register file or block.
    #   // Returns only the component of the HDL paths that corresponds to
    #   // the register file, not a full hierarchical path
    #   //
    #   // If no design abstraction is specified, the default design abstraction
    #   // for self register file is used.
    #   //
    #   extern def void get_hdl_path      (self,ref string paths[$], input string kind = ""):
    #
    #   //
    #   // Function:  get_full_hdl_path
    #   // Get the full hierarchical HDL path(s)
    #   //
    #   // Returns the full hierarchical HDL path(s) defined for the specified
    #   // design abstraction in the register file instance. If no design abstraction
    #   // is specified, uses the default design abstraction specified for the
    #   // nearest enclosing register file or block.
    #   // There may be more than one path returned even
    #   // if only one path was defined for the register file instance, if any of the
    #   // parent components have more than one path defined for the same design
    #   // abstraction
    #   //
    #   // If no design abstraction is specified, the default design abstraction
    #   // for each ancestor register file or block is used to get each
    #   // incremental path.
    #   //
    #   extern function void get_full_hdl_path (ref string paths[$],
    #                                           input string kind = "",
    #                                           input string separator = ".")
    #
    #   //
    #   // Function:    set_default_hdl_path
    #   // Set the default design abstraction
    #   //
    #   // Set the default design abstraction for self register file instance.
    #   //
    #   extern def void   set_default_hdl_path (self,string kind):
    #
    #   //
    #   // Function:  get_default_hdl_path
    #   // Get the default design abstraction
    #   //
    #   // Returns the default design abstraction for self register file instance.
    #   // If a default design abstraction has not been explicitly set for self
    #   // register file instance, returns the default design abstraction for the
    #   // nearest register file or block ancestor.
    #   // Returns "" if no default design abstraction has been specified.
    #   //
    #   extern def string get_default_hdl_path (self):
    #
    #
    #   extern virtual def void          do_print (self,uvm_printer printer):
    #   extern virtual def string        convert2string(self):
    #   extern virtual def uvm_object    clone      (self):
    #   extern virtual def void          do_copy    (self,uvm_object rhs):
    #   extern virtual function bit           do_compare (uvm_object  rhs,
    #                                                     uvm_comparer comparer)
    #   extern virtual def void          do_pack    (self,uvm_packer packer):
    #   extern virtual def void          do_unpack  (self,uvm_packer packer):
    #
    #endclass: uvm_reg_file
#
#
#//------------------------------------------------------------------------------
#// IMPLEMENTATION
#//------------------------------------------------------------------------------
#
#
#
#
#
#
#// get_regfile
#
#def uvm_reg_file uvm_reg_file::get_regfile(self):
#   return m_rf
#endfunction
#
#
#// clear_hdl_path
#
#def void uvm_reg_file::clear_hdl_path(self,string kind = "RTL"):
#  if (kind == "ALL"):
#    hdl_paths_pool = new("hdl_paths")
#    return
#  end
#
#  if (kind == ""):
#     if (m_rf is not None)
#        kind = m_rf.get_default_hdl_path()
#     else
#        kind = parent.get_default_hdl_path()
#  end
#
#  if (!hdl_paths_pool.exists(kind)):
#    `uvm_warning("RegModel",{"Unknown HDL Abstraction '",kind,"'"})
#    return
#  end
#
#  hdl_paths_pool.delete(kind)
#endfunction
#
#
#
#
#// has_hdl_path
#
#def bit  uvm_reg_file::has_hdl_path(self,string kind = ""):
#  if (kind == ""):
#     if (m_rf is not None)
#        kind = m_rf.get_default_hdl_path()
#     else
#        kind = parent.get_default_hdl_path()
#  end
#  
#  return hdl_paths_pool.exists(kind)
#endfunction
#
#
#// get_hdl_path
#
#def void uvm_reg_file::get_hdl_path(self,ref string paths[$], input string kind = ""):
#
#  uvm_queue #(string) hdl_paths
#
#  if (kind == ""):
#     if (m_rf is not None)
#        kind = m_rf.get_default_hdl_path()
#     else
#        kind = parent.get_default_hdl_path()
#  end
#
#  if (!has_hdl_path(kind)):
#    `uvm_error("RegModel",{"Register does not have hdl path defined for abstraction '",kind,"'"})
#    return
#  end
#
#  hdl_paths = hdl_paths_pool.get(kind)
#
#  for (int i=0; i<hdl_paths.size();i++)
#    paths.push_back(hdl_paths.get(i))
#
#endfunction
#
#
#// get_full_hdl_path
#
#function void uvm_reg_file::get_full_hdl_path(ref string paths[$],
#                                              input string kind = "",
#                                              input string separator = ".")
#   if (kind == "")
#      kind = get_default_hdl_path()
#
#   if (!has_hdl_path(kind)):
#      `uvm_error("RegModel",{"Register file does not have hdl path defined for abstraction '",kind,"'"})
#      return
#   end
#   
#   paths.delete()
#
#   begin
#      uvm_queue #(string) hdl_paths = hdl_paths_pool.get(kind)
#      string parent_paths[$]
#
#      if (m_rf is not None)
#         m_rf.get_full_hdl_path(parent_paths, kind, separator)
#      elif (parent is not None)
#         parent.get_full_hdl_path(parent_paths, kind, separator)
#
#      for (int i=0; i<hdl_paths.size();i++):
#         string hdl_path = hdl_paths.get(i)
#
#         if (parent_paths.size() == 0):
#            if (hdl_path != "")
#               paths.push_back(hdl_path)
#
#            continue
#         end
#         
#         foreach (parent_paths[j])  begin
#            if (hdl_path == "")
#               paths.push_back(parent_paths[j])
#            else
#               paths.push_back({ parent_paths[j], separator, hdl_path })
#         end
#      end
#   end
#
#endfunction
#
#
#// get_default_hdl_path
#
#def string uvm_reg_file::get_default_hdl_path(self):
#  if (default_hdl_path == ""):
#     if (m_rf is not None)
#        return m_rf.get_default_hdl_path()
#     else
#        return parent.get_default_hdl_path()
#  end
#  return default_hdl_path
#endfunction
#
#
#// set_default_hdl_path
#
#def void uvm_reg_file::set_default_hdl_path(self,string kind):
#
#  if (kind == ""):
#    if (m_rf is not None)
#       kind = m_rf.get_default_hdl_path()
#    elif (parent is None)
#       kind = parent.get_default_hdl_path()
#    else begin
#      `uvm_error("RegModel",{"Register file has no parent. ",
#           "Must specify a valid HDL abstraction (kind)"})
#      return
#    end
#  end
#
#  default_hdl_path = kind
#
#endfunction
#
#
#// get_parent
#
#def uvm_reg_block uvm_reg_file::get_parent(self):
#  return get_block()
#endfunction
#
#
#
#
#//-------------
#// STANDARD OPS
#//-------------
#
#// convert2string
#
#def string uvm_reg_file::convert2string(self):
#  `uvm_fatal("RegModel","RegModel register files cannot be converted to strings")
#   return ""
#endfunction: convert2string
#
#
#// do_print
#
#def void uvm_reg_file::do_print (self,uvm_printer printer):
#  super().do_print(printer)
#endfunction
#
#
#
#// clone
#
#def uvm_object uvm_reg_file::clone(self):
#  `uvm_fatal("RegModel","RegModel register files cannot be cloned")
#  return None
#endfunction
#
#// do_copy
#
#def void uvm_reg_file::do_copy(self,uvm_object rhs):
#  `uvm_fatal("RegModel","RegModel register files cannot be copied")
#endfunction
#
#
#// do_compare
#
#function bit uvm_reg_file::do_compare (uvm_object  rhs,
#                                        uvm_comparer comparer)
#  `uvm_warning("RegModel","RegModel register files cannot be compared")
#  return 0
#endfunction
#
#
#// do_pack
#
#def void uvm_reg_file::do_pack (self,uvm_packer packer):
#  `uvm_warning("RegModel","RegModel register files cannot be packed")
#endfunction
#
#
#// do_unpack
#
#def void uvm_reg_file::do_unpack (self,uvm_packer packer):
#  `uvm_warning("RegModel","RegModel register files cannot be unpacked")
#endfunction




#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.base import *
from uvm.comps import UVMAgent
from uvm.macros import uvm_component_utils, uvm_info

from ubus_slave_monitor import ubus_slave_monitor
from ubus_slave_driver import ubus_slave_driver
from ubus_slave_sequencer import ubus_slave_sequencer

#//------------------------------------------------------------------------------
#//
#// CLASS: ubus_slave_agent
#//
#//------------------------------------------------------------------------------


class ubus_slave_agent(UVMAgent):

    #  // constructor
    def __init__(self, name, parent):
        UVMAgent.__init__(self, name, parent)
        self.driver = None
        self.sequencer = None
        self.monitor = None
        self.tag = "UBUS_SLAVE_AGENT_" + name

    #  // build_phase
    def build_phase(self, phase):
        UVMAgent.build_phase(self, phase)
        self.monitor = ubus_slave_monitor.type_id.create("u_slv_monitor", self)
        if self.get_is_active() == UVM_ACTIVE:
            self.driver = ubus_slave_driver.type_id.create("u_slv_driver", self)
            self.sequencer = ubus_slave_sequencer.type_id.create("u_slv_sequencer",
                    self)

    # connect_phase
    def connect_phase(self, phase):
        if self.get_is_active() == UVM_ACTIVE:
            uvm_info(self.tag, "Connecting comps in active mode now",
                    UVM_MEDIUM)
            self.driver.seq_item_port.connect(self.sequencer.seq_item_export)
            self.sequencer.addr_ph_port.connect(self.monitor.addr_ph_imp)


uvm_component_utils(ubus_slave_agent)
